/*---------------------------------------------------------------------------*
Project:  Voice Chat Sample.
File:     net.h

Copyright (C) 2011-2014 Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain proprietary
information of Nintendo and/or its licensed developers and are protected by
national and international copyright laws. They may not be disclosed to third
parties or copied or duplicated in any form, in whole or in part, without the
prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#ifndef SAMPLE_NET_H_
#define SAMPLE_NET_H_

#include "definitions.h"

/*---------------------------------------------------------------------------*
    Include
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#include <nn.h>
#include <nn/ac.h>
#include <nn/socket_Initialize.h>
#include <nn/friends.h>
#else
#include<cafe.h>
#endif
#include <nex.h>
#include <pia/common.h>
#include <pia/session.h>
#ifdef VCT_PLATFORM_CTR
#include <pia/local.h>
#include <pia/local/local_UdsNetwork.h>
#endif
#include <pia/inet.h>

#include "utilities.h"

/*---------------------------------------------------------------------------*
    Define
 *---------------------------------------------------------------------------*/
#define NET_SEND_DATA_MAX  (80)
#define NET_SEND_BUFF_SIZE (2048)
#ifdef VCT_PLATFORM_CTR
#define NET_MAX_CONNECT    (8)
#else
#define NET_MAX_CONNECT    (16)
#endif


/*---------------------------------------------------------------------------*
    Enum
 *---------------------------------------------------------------------------*/
/* Network matchmake system mode. */
enum NET_MATCH_MODE
{
    NET_MATCH_ANYBODY = 0,  /* Matchmake is anybody mode.   */
    NET_MATCH_FRIEND,       /* Matchmake is friend mode.    */
    NET_MATCH_MODE_MAX      /* term...                      */
};
/* Network matchmake system mode. */
enum NET_EVENT_STATUS
{
    NET_EVENT_CONNECT = 0,      /* connect event.           */
    NET_EVENT_DISCONNECT,       /* disconnect event.        */
    NET_EVENT_CHANGE_CONNECTION /* Connection change event. */
};


/*---------------------------------------------------------------------------*
    Typedef
 *---------------------------------------------------------------------------*/
/* net information. */
typedef struct
{
    nn::pia::inet::SocketAddress hostAddr;        /* host address info.        */
    nn::pia::inet::SocketAddress netMask;         /* net mask info.            */
    nn::pia::inet::SocketAddress dns1;            /* 1st dns address info.     */
    nn::pia::inet::SocketAddress dns2;            /* 2nd dns address info.     */
    nn::pia::inet::SocketAddress gateway;         /* gateway address info.     */
    u32                CIDList[NET_MAX_CONNECT];  /* connect aid bitmap.       */
    u32                connectionNum;             /* connect num.              */
    u32                ownerPid;                  /* owner PID info.           */
    u32                myPid;                     /* my PID info.              */
    u16                myCid;                     /* my ConnectionID info.     */
#ifdef VCT_PLATFORM_CTR
    u8                 mac[nn::socket::ETH_ALEN]; /* mac address info.         */
#else
    //Cafe does not allow acquisition of MAC address.
#endif
} NetInformationStructure;

/*---------------------------------------------------------------------------*
  Name        : NetEventCallback
  Description : net event callback function.
  Arguments   : status - event status.
              : *info  - net information.
  Returns     : -
 *---------------------------------------------------------------------------*/
typedef void (*NetEventCallback)( NET_EVENT_STATUS status, NetInformationStructure *info );

/*---------------------------------------------------------------------------*
  Name        : NetReceiveCallback
  Description : net receive data callback function.
  Arguments   : srcCID  - src ConnectionID.
              : *buffer - received data.
              : size    - received data size.
  Returns     : -
 *---------------------------------------------------------------------------*/
typedef void (*NetReceiveCallback)( u32 srcCID, void *buffer, u32 size );

/* net config parameter. */
typedef struct
{
    NetEventCallback   eventCallbackFunction;   /* net event callback func.   */
    NetReceiveCallback receiveCallbackFunction; /* net receive callback func. */
    u32                gameMode;                /* net game mode.             */
    u32                maxConnect;              /* net max connect.           */
    NET_MATCH_MODE     netMatchMode;            /* net match mode.            */
} NetConfigStructure;
/* net send data information. */
typedef struct
{
    u32                   send_buff[NET_SEND_BUFF_SIZE/sizeof(u32)];
    nn::nex::ConnectionID destCid;           /* destination connection id info. */
    u32                   dataSize;          /* data size info.                 */
    bool                  reliable;          /* reliable send info.             */
} NetSendDataStructure;


/*---------------------------------------------------------------------------*
    Class
 *---------------------------------------------------------------------------*/
class PiaNetworkUtil
{
private:
    // This is the type of communication error to be handled by the sample program.
    enum SampleErrorType
    {
        NO_ERROR                      = 0, //
        INITIALIZE_ERROR_WIRELESS_OFF,     //
        INITIALIZE_ERROR_OTHER_REASON,     //
        MATCH_MAKE_ERROR,                  //
        SESSION_LOST_ERROR,                //
        CONNECTION_LOST_ERROR              //
    };
    static const u32 MEMORY_BLOCK_SIZE_PIA = 6 * 1024 * 1024;  // The size of the memory we give to Pia
    SampleErrorType m_ErrorType;                //Variable that maintains errors that occurred most recently
    nn::Result m_NetworkErrorResult;
    nn::nex::GatheringID m_JoinedGatheringId;   // Gathering ID to use for matchmaking
    nn::nex::PrincipalID m_OwnerPID;            // The host's principal ID
    nn::nex::PrincipalID m_MyPID;
    nn::nex::NgsFacade                *m_pNgsFacade;
    nn::nex::MatchmakeExtensionClient *m_pMatchMakingClient;
    u32 m_IsSessionMaster;
#ifdef VCT_PLATFORM_CTR
    bit8 m_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA] NN_ATTRIBUTE_ALIGN(4);
#else
    bit8 m_PiaMemoryBlock[MEMORY_BLOCK_SIZE_PIA] __attribute__ ((aligned(4)));
#endif
    void SetCommunicationErrorType(SampleErrorType);
    void SetNetworkErrorResult(nn::Result);
    nn::Result NexFacadeStartup(void);
    bool OpenParticipation(void);
public:
    bool InitializeCommon(void);
    void FinalizeCommon(void);
    bool InitializeInet(void);
    bool AutoMatch(u32,u32);
    bool Session(void);
    void FinalizeInet(void);
    nn::nex::PrincipalID GetOwnerPID(void) const {return m_OwnerPID;}
    nn::nex::PrincipalID GetMyPID(void) const {return m_MyPID;}
};

class NetworkInfra
{
public:
    NetworkInfra():m_PiaCommon(NULL){}
    bool Initialize(void);
    void Finalize(void);
    void Start( NetConfigStructure *config );
    void Stop( void );
    bool ReqSendData( u32 dstCID, void *buffer, u32 length, bool isReliable );
    bool GetNetInfo( NetInformationStructure *info ) const;
    bool CheckNetError( void );
    bool CheckConnection( void );
    void NetThreadImpl( void );
private:
    PiaNetworkUtil                    *m_PiaCommon;
    nn::nex::PrincipalID              m_MyPID;
    nn::nex::GatheringID              m_JoinedGatheringID;
    CriticalSection                   m_SendCC;
#ifdef VCT_PLATFORM_CTR
    nn::os::Thread                    m_NetThread;
#else
    OSThread                          *m_NetThread;
#endif
    u32                               m_NetThreadEndFlg;
    NetConfigStructure                m_Config;
    NetInformationStructure           m_Info;
    u32                               m_ReqSendDataIndxR;
    u32                               m_ReqSendDataIndxW;
    NetSendDataStructure              m_SendDataInfo[NET_SEND_DATA_MAX];
    bool                              m_SessionStartFlg;
    bool                              m_NetErrorFlg;
};


#endif /* SAMPLE_NET_H_ */
