/*---------------------------------------------------------------------------*
  Project:  Voice Chat Sample.
  File:     sndmic.h

  Copyright (C)1998-2014 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


#ifndef SAMPLE_SNDMIC_H_
#define SAMPLE_SNDMIC_H_

#include "definitions.h"

/*---------------------------------------------------------------------------*
 * Includes
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#include <nn.h>
#include <nn/os.h>
#include <nn/mic.h>
#include <nn/snd.h>
#include "framework.h"
#include "wave.h"
#else
#include <cafe/os.h>
#include <cafe/mic.h>
#include <cafe/ax.h>
#include <cafe/mix.h>
#include <cafe/usb_mic.h>
#endif
#include <vct.h>


/*---------------------------------------------------------------------------*
 *  Define
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#define SOUND_BUFFER_NUM        (3)         // Num of sound buffer
#else
#define SPEAKER_GAIN_IDX_MAX    (11)
#endif


/*---------------------------------------------------------------------------*
 *  Enum
 *---------------------------------------------------------------------------*/
/* mic & snd error status. */
typedef enum
{
    SNDMIC_ERROR_NONE = 0,          // None error
    SNDMIC_ERROR_PARAM,             // Parameter error
    MIC_ERROR_PARAM,                // Parameter error
    MIC_ERROR_SEQ,                  // Sequence error
    MIC_ERROR_SYS,                  // System error
    MIC_ERROR_LIB,                  // Library error
    MIC_WARN_FEW_DATA,              // Few data warning
    SND_ERROR_PARAM,                // Parameter error
    SND_ERROR_SEQ,                  // Sequence error
    SND_ERROR_SYS,                  // System error
    SND_ERROR_LIB,                  // SND library error
    SND_ERROR_DSP                   // DSP error
}SNDMIC_ERROR_STATUS;

/* mic process status. */
typedef enum
{
    SNDMIC_SEQ_NONE = 0,            // None process
    MIC_SEQ_INITED,                 // Initialize process
    MIC_SEQ_START,                  // Start process
    MIC_SEQ_STOP,                   // Stop process
    SND_SEQ_INITED,                 // Initialize process.
    SND_SEQ_START,                  // Start process.
    SND_SEQ_STOPPING,               // Stopping process.
    SND_SEQ_STOPPED                 // Stopped process.
}SNDMIC_SEQ_STATUS;

/* sampling freqency. */
typedef enum
{
    SND_FREQ_08KHZ = 0,             //  8kHz sampling.
    SND_FREQ_16KHZ,                 // 16kHz sampling.
    SND_FREQ_32KHZ,                 // 32kHz sampling.
    SND_FREQ_MAX                    // Terminater.
}SND_FREQ;

/* snd process status. */
typedef enum
{
    SND_PROCESS_NONE = 0,           // None process
    SND_PROCESS_INIT,               // Initialize process
    SND_PROCESS_PLAY,               // Play process
    SND_PROCESS_STOP                // Stop process
}SND_PROCESS_STATUS;

/* mic process status. */
typedef enum
{
    MIC_PROCESS_NONE = 0,           // None process
    MIC_PROCESS_INIT,               // Initialize process
    MIC_PROCESS_START,              // Start process
    MIC_PROCESS_STOP                // Stop process
}MIC_PROCESS_STATUS;

/*---------------------------------------------------------------------------*
 *  Struct
 *---------------------------------------------------------------------------*/
typedef struct
{
    unsigned int sampleRate;
    unsigned int micGainLevel;
    unsigned int micGainMin;
    unsigned int micGainMax;
    unsigned int micGainStep;
}MIC_STATE_INFO;


typedef struct
{
    SND_FREQ        freq;           // Sampling frequency.
    u32             frameSample;    // Sample of frame.
}SND_PARAM;


/* sndmic config parameter. */
typedef struct
{
    u32             frame_msec;    // Audio frame Length(mSec)
    u8              mic_gain;      // mic initialize gain
                                   /*
                                         -> 31 : MIC Gain(26.0dB, ~20)
                                            43 : MIC Gain(32.0dB, ~40)
                                            55 : MIC Gain(38.0dB, ~80)
                                            67 : MIC Gain(44.0dB, ~160)
                                   */
}SndMicConfigStructure;


/*---------------------------------------------------------------------------*
    Class
 *---------------------------------------------------------------------------*/
class SndMicDevice
{
#ifdef VCT_PLATFORM_CTR

public:
    // function
    SNDMIC_ERROR_STATUS     Initialize( SndMicConfigStructure *config );
    SNDMIC_ERROR_STATUS     Finalize( void );
    SNDMIC_ERROR_STATUS     Start( void );
    SNDMIC_ERROR_STATUS     Stop( void );
    SNDMIC_ERROR_STATUS     MicGainSet( u8 gain );
    SNDMIC_ERROR_STATUS     MicGainGet( u8 *gain ) const;
    SNDMIC_ERROR_STATUS     StartBGM(void);
    SNDMIC_ERROR_STATUS     StopBGM(void);
    SNDMIC_ERROR_STATUS     SetVolume( f32 volume );
    SNDMIC_ERROR_STATUS     SetBGMVolume( f32 volume );
    void                    GetMicData( void );
    void                    MicDataThreadImpl( void );
    void                    SoundThreadFuncImpl( void );
    void                    FrameProcessThreadImpl( void );
    bool                    GetSndMicErrorFlag( void);
    void                    RomfileWavePlay( u32 number );
    bool                    GetReceiveAudioData( u8 *audioBuf, u16 length );
    u8                      Pause( void );

private:
    // variable
    nn::os::Thread          m_SoundThread;
    nn::os::Thread          m_FrameProcessThread;
    nn::os::Event           m_MicSyncEvent;
    nn::os::Event           m_MicStopEvent;

    nn::snd::WaveBuffer     m_WaveBuffer[SOUND_BUFFER_NUM];
    nn::snd::WaveBuffer     m_MusicWaveBuffer;
    nn::snd::Voice*         m_pVoice;
    nn::snd::Voice*         m_pMusic;
    u8*                     m_SoundBuffer[SOUND_BUFFER_NUM];
    s16                     m_MixedData[NN_SND_SAMPLES_PER_FRAME * 2];
    s8                      m_SndBuf[4][VCT_AUDIO_DATA_SIZE_MAX];
    u16                     m_SndBufWriteIdx;
    u16                     m_SndBufReadIdx;
    SndMicConfigStructure   m_Config;
    u32                     m_SoundBufferSize;
    u32                     m_AudioFrameSize;
    u16                     m_CurrBufferIdx;
    u16                     m_SoundBufferReqIdx;
    SND_PROCESS_STATUS      m_SoundStatus;

    char                    m_MicBuffer[8192] NN_ATTRIBUTE_ALIGN(nn::mic::BUFFER_ALIGNMENT_ADDRESS);
    uptr                    m_MicBuffHalfPosition;
    u16                     m_MicBufferSize;
    u16                     m_MicPrevPosition;
    MIC_PROCESS_STATUS      m_MicStatus;
    u8                      m_MicGain;
    u8                      m_Pause;
    u8                      *m_pWaveMemory[3];
    u8                      *m_pPlayWave;
    u32                     m_FrameCount[3];
    u32                     m_FramePlayNum;
    u32                     m_WaveNumbber;
    WaveData                m_WaveData[3];
    bool                    m_FramePlayReq;
    bool                    m_MicTimingError;
    char                    m_zero_buf[192*8*2];
#else
public:
    SndMicDevice(){m_This = this;m_AttatchUsbMic = 0;}
    // mic function
    SNDMIC_ERROR_STATUS     InitializeMic( void );
    SNDMIC_ERROR_STATUS     FinalizeMic( void );
    SNDMIC_ERROR_STATUS     StartMic( void );
    SNDMIC_ERROR_STATUS     StopMic( void );
    SNDMIC_ERROR_STATUS     ReadMic( s16 *out_micData, s16 in_sample );
    SNDMIC_ERROR_STATUS     GetStateMic( MIC_STATE_INFO *out_state ) const;
    SNDMIC_ERROR_STATUS     SetGainLevelMic( unsigned int in_gainLevel );
    SNDMIC_ERROR_STATUS     GetGainLevelMic( unsigned int *out_gainLevel ) const;
    //snd function
    SNDMIC_ERROR_STATUS     InitializeSnd( const SND_PARAM *in_sndParam );
    SNDMIC_ERROR_STATUS     FinalizeSnd( void );
    SNDMIC_ERROR_STATUS     StartSnd( void );
    SNDMIC_ERROR_STATUS     StopSnd( void );
    SNDMIC_ERROR_STATUS     StartEffect( u16 in_effect );
    SNDMIC_ERROR_STATUS     StoreDataSnd( const s16 *in_buffer, s16 in_sample );
    SNDMIC_ERROR_STATUS     GetDRCGainSnd( u16 *out_drcGain ) const;
    SNDMIC_ERROR_STATUS     SetDRCGainSnd( u16 in_drcGain);
    SNDMIC_ERROR_STATUS     SetDRCVolume( u32 index );
    //common
    void                    InitializeMicSndCommon( SND_PARAM* );
    void                    FinalizeMicSndCommon(void);
    void                    InitializeUSBMIC( void );
    s32                     RSMP32to8(s16 *input, s32 num, s16* output);
    s32                     RSMP32to16(s16 *input, s32 num, s16* output);
    int                     MicSndFrameThreadImpl( void );
    u32                     Pause( SND_PARAM* );
    void                    AttatchUsbMicFlag(u32 n,usb_mic_handle_t h){m_AttatchUsbMic=n;m_UsbHandle=h;}
    static SndMicDevice     *m_This;
    static void             axFrameCallback( void );
private:
    void                    axFrameCallbackImpl( void );
    void                    axSetData( void );
    void                    readVoiceData( s16 *out_data, s16 in_sample );
    s32                     readFileData(const char *,FSClient *hFSClient,FSCmdBlock *hFSCmdBlock,s16 **buffer);
private:
    // variable
    mic_handle_t            m_hMic;
    SNDMIC_SEQ_STATUS       m_micSeq;
    s16                     *m_micRingBuffer;
    unsigned int            m_micSampleRate;
    unsigned int            m_micGainMin;
    unsigned int            m_micGainMax;
    unsigned int            m_micGainStep;
    unsigned int            m_micGainLevel;

    SNDMIC_SEQ_STATUS       m_sndSeq;
    AXVPB                   *m_voice;
    s16                     *m_streamBuffer;
    s16                     *m_voiceBuffer;
    u16                     m_writeIdx;
    u16                     m_maxIdx;
    u16                     m_blkSize;
    u16                     m_voiceWriteIdx;
    u16                     m_voiceReadIdx;
    u16                     m_voiceAvailable;
    u32                     m_frameSample;
    u32                     m_prevPtr;
    u32                     m_totalSample;
    SND_FREQ                m_freq;

    AXVPB                   *m_busy;
    AXVPB                   *m_noise;
    s16                     *m_busyBuffer;
    s16                     *m_noiseBuffer;

    OSEvent                 m_AudioFrameEvent;
    OSThread                *m_FrameThread;
    BOOL                    m_FrameThreadEndFlg;
    u32                     m_Pause;
    usb_mic_handle_t        m_UsbHandle;
    mem_res_element_t       *m_UsbResBuffer;
    u8                      m_AttatchUsbMic;
    u8                      m_UseUsbMic;
    u16                     m_dummy00;
#endif
};


#endif  /* SAMPLE_SNDMIC_H_ */
/* EOF */
