﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Inet_Login.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "Shared_Inet_Login.h"
#if NN_PIA_CTR
#include <nn/friends.h>
#endif

namespace {
    // The host's principal ID
    nn::nex::PrincipalID s_OwnerPID = INVALID_PRINCIPALID;
    // Login network error code.
    s32 s_LoginNetworkErrorCode = 0;
};

//-----------------------------------------------------------------------------
// Event handler to process notifications from the NGS
//-----------------------------------------------------------------------------
class SampleNotificationEventHandler : public nn::nex::NotificationEventHandler {
public:
    virtual void ProcessNotificationEvent(const nn::nex::NotificationEvent &oEvent) {
        nn::nex::StringStream ss;
        ss << oEvent;
        PIA_SAMPLE_LOG("ProcessNotificationEvent:%ls\n", ss.CStr());

        // Per-message processing
        switch (oEvent.GetType()) {
        case nn::nex::NotificationEvents::ParticipationEvent:
            switch (oEvent.GetSubType()) {
            case nn::nex::ParticipationEvents::Participate:
                PIA_SAMPLE_LOG("ParticipationEvents::Participate: [PID %u] join [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::EndParticipation:
                PIA_SAMPLE_LOG("ParticipationEvents::EndParticipation: [PID %u] end [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::Disconnect:
                PIA_SAMPLE_LOG("ParticipationEvents::Disconnect: [PID %u] disconnect [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            default:
                PIA_SAMPLE_LOG("ParticipationEvents: unsupported participation event:%d\n", oEvent.GetSubType());
                break;
            }
            break;

        case nn::nex::NotificationEvents::OwnershipChangeEvent:
            PIA_SAMPLE_LOG(
                "NotificationEvents::OwnershipChangeEvent: [session %u] changed owner to [PID %u]\n",
                oEvent.GetParam1(), 
                oEvent.GetParam2()
            );

            // Record the new owner's PID
            s_OwnerPID = oEvent.GetParam2();
            break;

        case nn::nex::NotificationEvents::GatheringUnregistered:
        {
            // Record the deleted gathering's GID
            nn::nex::GatheringID unregisteredGatheringId = oEvent.GetParam1();
            PIA_SAMPLE_LOG("NotificationEvents::GatheringUnregistered: [session %u] unregistered\n", unregisteredGatheringId);
            NN_PIA_DUMMY_PARAM(unregisteredGatheringId);
            
            break;
        }
        default:
            PIA_SAMPLE_LOG("NotificationEvents: unsupported notification event:%d\n", oEvent.GetType());
            break;
        }
    }
};

// Event handler to process notifications
static SampleNotificationEventHandler s_NotificationEventHandler;

//-----------------------------------------------------------------------------
// Log into the game server
// Logs onto the friend server at the same time
//-----------------------------------------------------------------------------
nn::Result Inet_Login(
    nn::nex::NgsFacade* pNgsFacade,
    u32 gameId, 
    const qChar* pAccessKey,
    u32 timeout
){
    nn::Result result;

    PIA_SAMPLE_LOG("Inet_Login\n");

#if ENABLE_LAN_NETWORK && NN_PIA_NEX_3
    PIA_SAMPLE_LOG("Inet_Login:using LAN network\n");

    //Take appropriate account name
    static const int NAMELEN = 20;
    static qChar buffer[NAMELEN+1]={'\0'};
    if (buffer[0] == 0) {
        for(int i = 0; i < NAMELEN; i++){
            int rndNum = nn::nex::Platform::GetRandomNumber(52);
            bool isBig = (rndNum < 26);
            buffer[i] =  ((isBig ? 'A' : 'a') + static_cast<nn::nex::qChar>(rndNum % 26));
        }
        buffer[NAMELEN] = '\0';
    }

    nn::nex::StringStream accountTemplateStream;
    accountTemplateStream << buffer;

    //Information for server to connect to
    //n2788: Address of matchmaking server inside LAN for P2P team.
    static qChar *pHost = _T("192.149.23.200");
    //static qChar *pHost = _T("localhost");
    static qUnsignedInt16 port = 9110;
    static qChar *pTestAccessKey = _T("Test");
    static qChar8 *pPassword = "password";

    //Create account
    {
        nn::nex::ProtocolCallContext pccCreate;
        nn::nex::qResult ret = pNgsFacade->DebugCreateAccount(
            &pccCreate, pHost, port, accountTemplateStream.CStr(),
            pPassword, pTestAccessKey, 30000 );

        if ( !ret )
        {
            PIA_SAMPLE_LOG("NgsFacade::DebugCreateAccount() call failed!!!\n");
            return SampleFailedNexLogin();
        }

        pccCreate.Wait();
        if ( pccCreate.GetState() != nn::nex::CallContext::CallSuccess )
        {
            PIA_SAMPLE_LOG("NgsFacade::DebugCreateAccount() failed.\n");
            //return NULL;
        }
    }

    //n2788: Provision for warnings about unused parameters.
    (void)&gameId;
    (void)&pAccessKey;

    nn::nex::ProtocolCallContext oContext;
    if (
        !pNgsFacade->DebugLoginDirect(
            &oContext, 
            pHost, 
            port, 
            accountTemplateStream.CStr(),
            pPassword, 
            pTestAccessKey, 
            nn::nex::TimeInterval(timeout)) 
       ) {
        // Failed to start the login process
        PIA_SAMPLE_LOG("Inet_Login:Login failed\n");
        return SampleFailedNexLogin();
    }
#else
#if NN_PIA_CTR
    nn::os::Event event(true);

    // Initialize friend library
    result = nn::friends::Initialize();
    if (result.IsFailure()) {
        // Can fail due to the internal state of the daemon
        PIA_SAMPLE_LOG("Inet_Login:friends::Initialize failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedFriendsLogin();
    }

    // Log into the friend server
    result = nn::friends::Login(&event);
    if (result.IsFailure()) {
        // Can fail if not connected to an AP
        PIA_SAMPLE_LOG("Inet_Login:friends::Login failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedFriendsLogin();
    }

    // Wait for asynchronous login processing to end
    // Put code here for readability.
    event.Wait();

    // Get results of asynchronous login process
    result = nn::friends::GetLastResponseResult();
    if (result.IsFailure()) {
        // Can fail due to the internal state of the daemon
        PIA_SAMPLE_LOG("Inet_Login:friends::GetLastResponseResult failed:ErrorCode:%d\n", nn::friends::ResultToErrorCode(result));
        PIA_TRACE_RESULT(result);
        return SampleFailedFriendsLogin();
    }

    PIA_SAMPLE_LOG("Inet_Login:friends::Login succeeded:MyPrincipalId:%d\n", nn::friends::GetMyPrincipalId());

    // Once you're able to log in to the friend server, next, log in to the game server
    if (!nn::friends::HasLoggedIn()) {
        PIA_SAMPLE_LOG("Inet_Login:HasLoggedIn failed\n");
        return SampleFailedFriendsLogin();
    }
#elif NN_PIA_CAFE
    // Do not need to log in to friend server.
#endif

    PIA_SAMPLE_LOG("Logging in to Game Server...\n");

    // Log into the game server
    nn::nex::ProtocolCallContext oContext;
    if (
        !pNgsFacade->Login(
            &oContext, 
            gameId, 
            pAccessKey, 
            nn::nex::TimeInterval(timeout)
        )
    ) {
        // Failed to start the login process
        PIA_SAMPLE_LOG("Inet_Login:Login failed\n");
        return SampleFailedNexLogin();
    }
#endif // ENABLE_LAN_NETWORK

    // Wait for asynchronous login processing to end
    // nn::nex::Scheduler::Dispatch is called in ProtocolCallContext::Wait, and the asynchronous processing progresses.
    // Put code here for readability.
    oContext.Wait();
    if (!oContext.GetOutcome()) {
        PIA_SAMPLE_LOG("Inet_Login:Login failed. NetworkErrorCode: %d, LoginErrorCode: %d\n", 
           nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(oContext.GetOutcome()), pNgsFacade->GetLastLoginErrorCode());
        s_LoginNetworkErrorCode = pNgsFacade->GetLastLoginErrorCode();
        return SampleFailedNexLogin();
    }

    if (!pNgsFacade->IsConnected()) {
        PIA_SAMPLE_LOG("Inet_Login:Login failed. LoginErrorCode: %d\n", pNgsFacade->GetLastLoginErrorCode());
        s_LoginNetworkErrorCode = pNgsFacade->GetLastLoginErrorCode();
        return SampleFailedNexLogin();
    }

    // Register an event handler to process notifications from the NGS
    pNgsFacade->RegisterNotificationEventHandler(&s_NotificationEventHandler);

    PIA_SAMPLE_LOG("Inet_Login:Login succeeded\n");
    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------
// Logs out from game server
//-----------------------------------------------------------------------------
nn::Result Inet_Logout(nn::nex::NgsFacade* pNgsFacade) {
    // Delete the event handler
    pNgsFacade->UnregisterNotificationEventHandler(&s_NotificationEventHandler);

    // Logging out from the game server
    nn::nex::ProtocolCallContext ccLogout;
    PIA_SAMPLE_LOG("Inet_Logout:Logout\n");
    pNgsFacade->Logout(&ccLogout);
    ccLogout.Wait();

#if ENABLE_LAN_NETWORK
#else
#if NN_PIA_CTR
    PIA_SAMPLE_LOG("Inet_Logout:friends::Logout\n");
    // Logging out from the friend server
    nn::friends::Logout();

    // Terminate friend library
    nn::friends::Finalize();
#elif NN_PIA_CAFE
    // Do nothing because not logged in to friend server.
#endif
#endif

    PIA_SAMPLE_LOG("Inet_Logout:Logout succeeded\n");
    return nn::ResultSuccess();
}

nn::nex::PrincipalID Inet_GetOwnerPID() {
    return s_OwnerPID;
}

void Inet_SetOwnerPID(nn::nex::PrincipalID id) {
    s_OwnerPID = id;
}

s32 Inet_GetLoginNetworkErrorCode() {
    return s_LoginNetworkErrorCode;
}

void Inet_ResetLoginNetworkErrorCode() {
    s_LoginNetworkErrorCode = 0;
}
