/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Inet_Matchmake.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "Shared_Inet_Matchmake.h"
#include "Shared_Inet_Login.h"
#include "Shared_Nex_Setup.h"

// Index of the search attribute to use
#define MATCHMAKE_ATTR_INDEX 0

// The maximum search attribute value to use
#define MATCHMAKE_ATTR_MAX 2

namespace {
    // Matchmaking network error code.
    u32 s_MatchmakeNetworkErrorCode = 0;
    // Size of signature key configured in SignatureSetting object.
    const u32 cSignatureKeyDataSize = 32;
    // Settings for signature feature to use a shared session key in a PiaSession object.
    nn::pia::common::SignatureSettingWithKeyBuffer<cSignatureKeyDataSize> s_SignatureSetting;
};

// Define function to access shared session key used to manage element in the Shared_Nex_Setup function.
nn::nex::qVector<nn::nex::qByte>* GetMatchmakeSessionKeyPtr();


//n1589: todo2: API name to the log
//-----------------------------------------------------------------------------
// Run automatic matchmaking
// Here we get the necessary gathering and then join or create a session
//-----------------------------------------------------------------------------
nn::Result Inet_AutoMatchmake(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
    nn::nex::GatheringID* pJoinedGatheringID,
    nn::nex::PrincipalID* pOwnerPID,
    u16 matchmakeMaxParticipants,
    u32 matchmakeMode,
    const char* matchmakeDescription
){
    PIA_SAMPLE_LOG("Inet_AutoMatchmake:matchmakeMode:%d, matchmakeMaxParticipants:%d\n", matchmakeMode, matchmakeMaxParticipants);

    // Create search criteria
    nn::nex::qList<nn::nex::MatchmakeSessionSearchCriteria> criteriaList;

    // The first criteria uses a perfect match
    {
        nn::nex::MatchmakeSessionSearchCriteria criteria;

        // The minimum and maximum number of participants
        criteria.SetMinParticipants(1);
        criteria.SetMaxParticipants(matchmakeMaxParticipants);

        // Game mode
        criteria.SetGameMode(matchmakeMode);

        // Search attributes that can be freely defined by the application. Originally fixed at 0.
        criteria.SetAttribute(MATCHMAKE_ATTR_INDEX, 0);

        // Matchmaking system type
        //criteria.SetMatchmakeSystemType(netMatchMode);

        // Register to list
        criteriaList.push_back(criteria);
    }

    // The second criteria is specified within a range
    {
        nn::nex::MatchmakeSessionSearchCriteria criteria;

        // The minimum and maximum number of participants
        criteria.SetMinParticipants(1);
        criteria.SetMaxParticipants(matchmakeMaxParticipants);

        // Game mode
        criteria.SetGameMode(matchmakeMode);

        // Search attributes that can be freely defined by the application. All ranges that the attributes can take.
        criteria.SetAttributeWithRange(MATCHMAKE_ATTR_INDEX, 0, MATCHMAKE_ATTR_MAX);
        criteria.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

        // Register the criteria with the list (add it to the end of the list because it is the second search criteria)
        criteriaList.push_back(criteria);
    }

// Room created when no session matches the criteria
    nn::nex::MatchmakeSession* pMatchmakeSession = qNew nn::nex::MatchmakeSession();
    pMatchmakeSession->SetOpenParticipation(false);

    // The MatchmakeSession's description string setting is optional
    if ( matchmakeDescription == NULL ) 
    {
        pMatchmakeSession->SetDescription(_T("PiaInet Automatch"));
    }
    else
    {
        pMatchmakeSession->SetDescription(matchmakeDescription);
    }

    // The minimum and maximum number of participants
    pMatchmakeSession->SetMinParticipants(1);
    pMatchmakeSession->SetMaxParticipants(matchmakeMaxParticipants);

    // Game mode
    pMatchmakeSession->SetGameMode(matchmakeMode);

    // Search attributes that can be freely defined by the application
    // Because this is a sample, randomize attribute values for sessions that we create on our own
    pMatchmakeSession->SetAttribute(MATCHMAKE_ATTR_INDEX, nn::nex::Platform::GetRandomNumber(MATCHMAKE_ATTR_MAX + 1));

    // Matchmaking system type
    pMatchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Binary data that can be freely set by the application
    nn::nex::qVector<nn::nex::qUnsignedInt8> applicationBuffer;
    applicationBuffer.push_back(1);
    applicationBuffer.push_back(2);
    applicationBuffer.push_back(3);
    pMatchmakeSession->SetApplicationBuffer(applicationBuffer);

    // Transfer ownership when the owner logs out?
    pMatchmakeSession->SetFlag(nn::nex::GatheringFlags::MigrateOwner);

// The MatchmakeSession instance is destroyed by the InGatheringHolder destructor.
    nn::nex::GatheringHolder oInGatheringHolder = pMatchmakeSession;
    nn::nex::GatheringHolder oOutGatheringHolder;   // The gathering that we have joined
    
    // Search for or create a matchmaking session
    nn::nex::ProtocolCallContext oContext;

// Due to AutoMatchmake specifications, when leaving a matchmake session, that session cannot be found subsequently.
    bool bSuccess = pMatchMakingClient->AutoMatchmake(
        &oContext, 
        criteriaList, 
        oInGatheringHolder, 
        &oOutGatheringHolder, 
        _T("AutoMatchmake message")
    );

    if (!bSuccess) {
        PIA_SAMPLE_LOG("Failed to AutoMatchmake\n");
        return SampleFailedAutoMatchmakeSession();
    }

    oContext.Wait();
    if (oContext.GetState() != nn::nex::CallContext::CallSuccess) {
        Nex_ReportProtocolCallContext(oContext, "Failed to Inet_AutoMatchmake\n");
        return SampleFailedAutoMatchmakeSession();
    }

    *pJoinedGatheringID = oOutGatheringHolder->GetID();
    *pOwnerPID = oOutGatheringHolder->GetOwnerPID();

    nn::nex::qVector<nn::nex::qByte>* pMatchmakeSessionKey = GetMatchmakeSessionKeyPtr();
    pMatchmakeSessionKey->clear();
#if NN_PIA_NEX_3
    //Get the key sent by the game server as the signature key for the P2P session.
    *(pMatchmakeSessionKey) = reinterpret_cast<nn::nex::MatchmakeSession*>(oOutGatheringHolder.Get())->GetSessionKey();
#else
    //Use the gathering ID value repeated several times as the signature key.
    for (int i=0; i<8; i++) {
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >> 24) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >> 16) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >>  8) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >>  0) & 0xff);
    }
#endif

    PIA_SAMPLE_LOG(
        "[PID Info >> JoinedGID:%d OwnerPID:%d HostPID:%d MyPID:%d]\n", 
        oOutGatheringHolder->GetID(), 
        oOutGatheringHolder->GetOwnerPID(), 
        oOutGatheringHolder->GetHostPID(), 
        pMatchMakingClient->GetCredentials()->GetPrincipalID()
    );

    return nn::ResultSuccess();
}


//-----------------------------------------------------------------------------
// Open a matchmaking session
//-----------------------------------------------------------------------------
nn::Result Inet_OpenParticipation(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient, 
    nn::nex::GatheringID gatheringID
){
    nn::nex::ProtocolCallContext oContext;
    if (pMatchMakingClient->OpenParticipation(&oContext, gatheringID)) {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess) {
            Nex_ReportProtocolCallContext(oContext, "Inet_OpenParticipation:Failed to open participation\n");
            return SampleFailedOpenParticipation();
        }
    }
    else {
        PIA_SAMPLE_LOG("Inet_OpenParticipation:OpenParticipation failed (Startup)\n");
        return SampleFailedOpenParticipation();
    }
    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------
// Leave the matchmaking session
//-----------------------------------------------------------------------------
nn::Result Inet_LeaveMatchmakeSession(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient, 
    nn::nex::GatheringID gatheringID,
    nn::nex::PrincipalID ownerPID,
    bool hostMigration
){
    PIA_SAMPLE_LOG("Inet_LeaveMatchmakeSession\n");

    // Check the connection status with the mcs server.
    if( !pMatchMakingClient->IsConnected() )
    {
        PIA_SAMPLE_LOG("Already disconnected to server.\n");
        return SampleResultInvalidState();
    }

    nn::nex::ProtocolCallContext oContext;

    PIA_SAMPLE_LOG(
        "[PID Info >> OwnerPID:%d MyPID:%d]\n", 
        ownerPID, 
        pMatchMakingClient->GetCredentials()->GetPrincipalID()
    );

    // If a matchmaking session and if no host migration, deletes the matchmnaking session.
    if (pMatchMakingClient->GetCredentials()->GetPrincipalID() == ownerPID && !hostMigration) {
        PIA_SAMPLE_LOG("Unregister gathering ... \n");
        if (!pMatchMakingClient->UnregisterGathering(&oContext, gatheringID)) {
            PIA_SAMPLE_LOG("UnregisterGathering call failed ... \n");
            return SampleResultInvalidState();
        }

        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess) {
            Nex_ReportProtocolCallContext(oContext, "failed to Unregister gathering!!\n");
            return SampleResultInvalidState();
        }
        PIA_SAMPLE_LOG("Unregister gathering successfly.\n");
    }
    // If host migration is due to the owner or if not the owner, exits the matchmaking session.
    else {
        PIA_SAMPLE_LOG("End participation ... \n");
        if (!pMatchMakingClient->EndParticipation(&oContext, gatheringID)) {
            PIA_SAMPLE_LOG("Failed to end participation\n");
            return SampleResultInvalidState();
        }

        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess) {
            // If it has already been deleted, treat this as a normal exit.
            if (oContext.GetOutcome() == QRESULT_ERROR(RendezVous, SessionVoid)) {
                PIA_SAMPLE_LOG("matchmake session is already unregistered. don't care\n");
                return nn::ResultSuccess();
            }

            Nex_ReportProtocolCallContext(oContext, "Failed to end participation\n");
            return SampleResultInvalidState();
        }
    }
    
    return nn::ResultSuccess();
}

nn::Result Inet_GetHostStationURLs(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
    nn::nex::GatheringID joinedGatheringID,
    nn::nex::qList<nn::nex::StationURL>* pStationURLs
){
    // List the session URLs and try to join them
    nn::nex::ProtocolCallContext oContext;
    if (!pMatchMakingClient->GetSessionURLs(&oContext, joinedGatheringID, pStationURLs))
    {
        PIA_SAMPLE_LOG("Failed Request URLs...\n");
        return SampleResultInvalidState();
    }

    oContext.Wait();
    if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
    {
        // We may have joined a matchmaking session that has been deleted.
        if(oContext.GetOutcome() == QRESULT_ERROR(RendezVous, SessionVoid))
        {
            PIA_SAMPLE_LOG("matchmake session is already unregistered.\n");
        }
        Nex_ReportProtocolCallContext(oContext, "Failed to obtain a list of URLs to connect to the game host\n");
        return SampleResultInvalidState();
    }

    return nn::ResultSuccess();
}


//-----------------------------------------------------------------------------
// Searching for a Matchmaking Session
//-----------------------------------------------------------------------------
nn::Result Inet_SearchMatchmakeSession(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
    u32 matchmakeMode, 
    u16 matchmakeMaxParticipants,
    u32 rangeSize,
    nn::nex::qList<nn::nex::GatheringHolder>* plistGatheringHolder
){
    PIA_CACHED_PRINTF("Inet_SearchMatchmakeSession:start.\n");

    /* Start search condition settings */
    // See the CTR-NEX Function Reference or other documents for details on each setting.
    nn::nex::MatchmakeSessionSearchCriteria criteria;

    // Minimum/Maximum Number of Participants
    criteria.SetMinParticipants(1);
    criteria.SetMaxParticipants(matchmakeMaxParticipants);

    // Application-specific game mode.
    criteria.SetGameMode(matchmakeMode);

    // Search attributes that can be freely defined by the application (version where the range is specified)
    criteria.SetAttributeWithRange(MATCHMAKE_ATTR_INDEX, 0, MATCHMAKE_ATTR_MAX);

    // Matchmaking system type
    criteria.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Search only for sessions that can be joined
    criteria.SetExcludeLocked(true);

    // Exclude full sessions
    criteria.SetVacantOnly(true);
    /* Terminates search criteria settings */

    // Search Range
    nn::nex::ResultRange range;
    range.SetSize(rangeSize);   // Maximum number obtained
    range.SetOffset(0);         // Offset value

    // Empties the save list
    plistGatheringHolder->clear();

    nn::nex::ProtocolCallContext searchContext;
    nn::Result result;

    // Start asynchronous search process
    if( !pMatchMakingClient->BrowseMatchmakeSession(&searchContext, criteria, range, plistGatheringHolder) )
    {
        // Asynchronous search process failed to start.
        PIA_CACHED_PRINTF("Inet_SearchMatchmakeSession:BrowseMatchmakeSession failed.(Startup)\n");
        result = SampleResultInvalidState();
    }
    else
    {
        // Wait for asynchronous search process to end
        searchContext.Wait();
        if( !searchContext.GetOutcome() )
        {
            // Asynchronous search process failed
            PIA_CACHED_PRINTF("Inet_SearchMatchmakeSession:BrowseMatchmakeSession failed.(Search)\n");

            // Empties the save list
            plistGatheringHolder->clear();
            result = SampleResultInvalidState();
        }
        else
        {
            result = nn::ResultSuccess();
        }
    }

    PIA_CACHED_PRINTF("Inet_SearchMatchmakeSession:end.\n");
    return result;
}

//-----------------------------------------------------------------------------
// Opens a matchmaking session
//-----------------------------------------------------------------------------
nn::Result Inet_CreateMatchmakeSession(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
    nn::nex::GatheringID* pJoinedGatheringID,
    u32 matchmakeMode, 
    u16 matchmakeMaxParticipants,
    const char* matchmakeDescription
){
    // Create a new session.
    nn::nex::MatchmakeSession* pMatchmakeSession = qNew nn::nex::MatchmakeSession();

    // Matchmaking session settings

    // The MatchmakeSession's description string setting is optional
    pMatchmakeSession->SetDescription(matchmakeDescription);

    // The minimum and maximum number of participants
    pMatchmakeSession->SetMinParticipants(1);
    pMatchmakeSession->SetMaxParticipants(matchmakeMaxParticipants);

    // Application-specific game mode.
    pMatchmakeSession->SetGameMode(matchmakeMode);

    // Search attributes that can be freely defined by the application
    // As a sample, the attribute values are randomly set within the specification range.
    pMatchmakeSession->SetAttribute(MATCHMAKE_ATTR_INDEX, nn::nex::Platform::GetRandomNumber(MATCHMAKE_ATTR_MAX + 1));

    // Matchmaking system type
    pMatchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Binary data that can be freely set by the application
    // As a sample, a suitable value is set.
    nn::nex::qVector<nn::nex::qUnsignedInt8> applicationBuffer;
    applicationBuffer.push_back(1);
    applicationBuffer.push_back(2);
    applicationBuffer.push_back(3);
    pMatchmakeSession->SetApplicationBuffer(applicationBuffer);

    // Transfer ownership when the owner logs out?
    pMatchmakeSession->SetFlag(nn::nex::GatheringFlags::MigrateOwner);

    // Whether to set to joinable from the moment the session was created
    // Here, set to false, and use OpenParticipation after the Session (P2P communication) preparations are done.
    pMatchmakeSession->SetOpenParticipation(false);

    // The MatchmakeSession instance is destroyed by the gatheringHolder destructor.
    nn::nex::GatheringHolder gatheringHolder(pMatchmakeSession);
    nn::nex::ProtocolCallContext createContext;

    nn::nex::qVector<nn::nex::qByte>* pMatchmakeSessionKey = GetMatchmakeSessionKeyPtr();
    pMatchmakeSessionKey->clear();
    // Start asynchronous matchmaking session creation process
#if NN_PIA_NEX_3
    if( pMatchMakingClient->CreateMatchmakeSession(&createContext, gatheringHolder, pJoinedGatheringID, pMatchmakeSessionKey) ) {
#else
    if( pMatchMakingClient->CreateMatchmakeSession(&createContext, gatheringHolder, pJoinedGatheringID) ) {
#endif
        // Wait for the asynchronous matchmaking session creation process to end
        createContext.Wait();
        if( !createContext.GetOutcome() )
        {
            // Asynchronous matchmaking session creation process failed
            return SampleFailedCreateMatchmakeSession();
        }
    }
    else
    {
        // Asynchronous matchmaking session creation process failed to start
        return SampleFailedCreateMatchmakeSession();
    }

#if NN_PIA_NEX_3
#else
    //Use the gathering ID value repeated several times as the signature key.
    for (int i=0; i<8; i++) {
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >> 24) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >> 16) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >>  8) & 0xff);
        pMatchmakeSessionKey->push_back((*pJoinedGatheringID >>  0) & 0xff);
    }
#endif

    PIA_SAMPLE_LOG(
        "[PID Info >> JoinedGID:%d OwnerPID:%d HostPID:%d MyPID:%d]\n", 
        gatheringHolder->GetID(), 
        gatheringHolder->GetOwnerPID(), 
        gatheringHolder->GetHostPID(), 
        pMatchMakingClient->GetCredentials()->GetPrincipalID()
    );
    
    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------
// Join a Matchmaking Session
//-----------------------------------------------------------------------------
nn::Result Inet_JoinMatchmakeSession(
    nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
    nn::nex::GatheringID gatheringID
){
    nn::nex::ProtocolCallContext joinContext;

    nn::nex::qVector<nn::nex::qByte>* pMatchmakeSessionKey = GetMatchmakeSessionKeyPtr();
    pMatchmakeSessionKey->clear();
    // Start asynchronous matchmaking session participation process
#if NN_PIA_NEX_3
    if( pMatchMakingClient->JoinMatchmakeSession(&joinContext, gatheringID, pMatchmakeSessionKey) )
#else
    if( pMatchMakingClient->JoinMatchmakeSession(&joinContext, gatheringID) )
#endif
    {
        // Wait for the asynchronous matchmaking session participation process to end
        joinContext.Wait();
        if( !joinContext.GetOutcome() )
        {
            // Asynchronous matchmaking session participation process failed
            // Handling of the failure reason
            if( joinContext.GetOutcome() == QRESULT_ERROR(RendezVous, SessionFull) )
            {
                PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (SessionFull)\n" );
            }
            else if( joinContext.GetOutcome() == QRESULT_ERROR(RendezVous, SessionVoid) )
            {
                PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (SessionVoid)\n" );
            }
            else
            {
                if( joinContext.GetOutcome() == QRESULT_ERROR(RendezVous, DeniedByParticipants) )
                {
                    PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (DeniedByParticipants)\n" );
                }
                else if( joinContext.GetOutcome() == QRESULT_ERROR(RendezVous, ParticipantInBlackList) )
                {
                    PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (ParticipantInBlackList)\n" );
                }
                else
                {
                    PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (Unknown)\n" );
                }
                s_MatchmakeNetworkErrorCode = nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(joinContext.GetOutcome());
            }
            return SampleFailedJoinMatchmakeSession();
        }
    }
    else
    {
        PIA_CACHED_PRINTF( "JoinMatchmakeSession Error (Startup)\n" );
        // Asynchronous matchmaking session participation process failed to start
        return SampleFailedJoinMatchmakeSession();
    }

#if NN_PIA_NEX_3
#else
    //Use the gathering ID value repeated several times as the signature key.
    for (int i=0; i<8; i++) {
        pMatchmakeSessionKey->push_back((gatheringID >> 24) & 0xff);
        pMatchmakeSessionKey->push_back((gatheringID >> 16) & 0xff);
        pMatchmakeSessionKey->push_back((gatheringID >>  8) & 0xff);
        pMatchmakeSessionKey->push_back((gatheringID >>  0) & 0xff);
    }
#endif

    return nn::ResultSuccess();
}

u32 Inet_GetMatchmakeNetworkErrorCode() {
    return s_MatchmakeNetworkErrorCode;
}

void Inet_ResetMatchmakeNetworkErrorCode() {
    s_MatchmakeNetworkErrorCode = 0;
}

nn::nex::qVector<nn::nex::qByte>* Inet_GetMatchmakeSessionSignatureKey() {
    return GetMatchmakeSessionKeyPtr();
}

nn::pia::common::SignatureSetting* Inet_GetMatchmakeSessionSignatureSetting() {
    nn::nex::qVector<nn::nex::qByte>* pMatchmakeSessionKey = GetMatchmakeSessionKeyPtr();
    if (pMatchmakeSessionKey == NULL)  { return NULL; }
    nn::Result result = nn::pia::inet::NexFacade::ConvertNexSignatureKeyToSignatureSetting(
        *(pMatchmakeSessionKey),
        &s_SignatureSetting
    );
    if (result.IsFailure()) {
        return NULL;
    }
    return &s_SignatureSetting;
}

