/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Inet_NetSetup.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "Shared_Inet_NetSetup.h"

#include <pia/common/common_String.h>
#include <pia/inet/inet_SocketAddress.h>

namespace {
#if NN_PIA_CTR
    /*
        When using nn::nex::Core::ThreadModeUnsafeTransportBuffer:
        you need a total of four buffers: two for NEX and two for Pia.

        When using nn::nex::Core::ThreadModeUnsafeUser:
        you need a total of three buffers: one for NEX and two for Pia.
    */
    #define SOCKET_SESSION_MAX 4
    #define SOCKET_BUFFER (64 * 1024)
    // nn::socket::Initialize requires a 4096-byte aligned buffer
    PIA_ATTRIBUTE_ALIGN(4096) u8 s_SocketBuffer[SOCKET_BUFFER * 2];

    // Event that fires when a disconnection notification is received.
    nn::os::Event s_ACDisconnectedEvent;
#elif NN_PIA_CAFE
    // Connection status flag.
    bool s_ACConnected = false;
    bool s_UpdateACStatusRun = false;
    // Thread and thread stack for periodically checking the connection status of the AC.
    PIA_ATTRIBUTE_ALIGN(8) Thread s_UpdateACStatusThread;
    PIA_ATTRIBUTE_ALIGN(4) u8 s_UpdateACStatusThreadStack[4 * 1024];

    int UpdateACStatus(int argc, void* argv)
    {
        NN_PIA_DUMMY_PARAM(argc);
        NN_PIA_DUMMY_PARAM(argv);
        
        while ( s_UpdateACStatusRun ) {
            bool IsConnected;
            nn::Result result = nn::ac::IsApplicationConnected(&IsConnected);
            if (result.IsSuccess())
            {
                // Save results so that they can be accessed from the main thread.
                s_ACConnected = IsConnected;
            }
            else
            {
                // Set status to "disconnected" if the nn::ac::Initialize function has not been called.
                s_ACConnected = false;
            }

            // This loops about every 100 msec.
            OSSleepMilliseconds(100);
        }
        return 0;
    }
#endif

    void PrintIfConfig() {
        nn::pia::inet::SocketAddress host, netmask, dns1, dns2, gateway;
        nn::pia::inet::SockAddrIn *pSockAddrIn;

#if NN_PIA_CTR
        s32 ret;
        nn::socket::InAddr addr1, addr2;

        NN_UNUSED_VAR(ret);

        ret = nn::socket::GetPrimaryAddress(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2)); 
        PIA_SAMPLE_ASSERT(ret == 0);
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(host.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr1.addr;
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(netmask.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr2.addr;

        ret = nn::socket::GetResolver(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2));
        PIA_SAMPLE_ASSERT(ret == 0);
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(dns1.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr1.addr;
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(dns2.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr2.addr;
        
        ret = nn::socket::GetDefaultGateway(reinterpret_cast<u8*>(&addr1));
        PIA_SAMPLE_ASSERT(ret == 0);
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(gateway.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr1.addr;
        
#elif NN_PIA_CAFE
        nn::Result result;
        nn::ac::IpAddress addr;

        result = nn::ac::GetAssignedAddress(&addr);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(host.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr;
        
        result = nn::ac::GetAssignedSubnet(&addr);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(netmask.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr;
        
        result = nn::ac::GetAssignedPreferedDns(&addr);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(dns1.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr;
        
        result = nn::ac::GetAssignedAlternativeDns(&addr);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(dns2.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr;
        
        result = nn::ac::GetAssignedGateway(&addr);
        PIA_SAMPLE_ASSERT(result.IsSuccess());
        pSockAddrIn = const_cast<nn::pia::inet::SockAddrIn*>(gateway.GetSockAddrInPointer());
        pSockAddrIn->addr.addr = addr;
#endif
        nn::pia::common::String      string;
        nn::pia::common::InetAddress inetAddress;

        host.GetInetAddress(&inetAddress);
        inetAddress.GetAddressString(&string);
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:host   : %s\n", string.GetCharArray());

        netmask.GetInetAddress(&inetAddress);
        inetAddress.GetAddressString(&string);
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:netmask: %s\n", string.GetCharArray());

        dns1.GetInetAddress(&inetAddress);
        inetAddress.GetAddressString(&string);
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:dns1   : %s\n", string.GetCharArray());

        dns2.GetInetAddress(&inetAddress);
        inetAddress.GetAddressString(&string);
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:dns2   : %s\n", string.GetCharArray());

        gateway.GetInetAddress(&inetAddress);
        inetAddress.GetAddressString(&string);
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:gateway: %s\n", string.GetCharArray());

#if NN_PIA_CTR
    #ifndef NN_SWITCH_DISABLE_DEBUG_PRINT
        PIA_SAMPLE_LOG("\n");
        nn::socket::DumpRoutingTable();
    #endif
#endif
    }
};



nn::Result Inet_InitializeNetwork() {
#if NN_PIA_CTR
    nn::Result result;
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(2000));

// Need to call fs::Initialize, but no need to call fs::Finalize
    nn::fs::Initialize();

    PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Initialize\n");
    result = nn::ac::Initialize();
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Initialize:failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }

    nn::ac::Config config;
    nn::ac::CreateDefaultConfig( &config );

#ifndef RELEASE_VERSION
    nn::ac::DebugSetNetworkArea( &config, nn::ac::NETWORK_AREA_AUTO );
#endif

    PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Connect:\n");
    result = nn::ac::Connect( config );
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Connect:failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }

    // Register event for receiving disconnection notifications.
    s_ACDisconnectedEvent.Initialize(true);
    nn::ac::CTR::RegisterDisconnectEvent(&s_ACDisconnectedEvent);

    // With NEX, socket APIs can be called from two threads.
    size_t requiredMemorySize = nn::socket::GetRequiredMemorySize(SOCKET_BUFFER, SOCKET_SESSION_MAX);
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:socket::Initialize\n");
    result = nn::socket::Initialize(reinterpret_cast<uptr>(s_SocketBuffer), requiredMemorySize, SOCKET_BUFFER, SOCKET_SESSION_MAX);
    PIA_SAMPLE_ASSERT(requiredMemorySize < sizeof(s_SocketBuffer));
    
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:socket::Initialize:failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }
#elif NN_PIA_CAFE
    nn::Result result;

    // Initialize the Account library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:act::Initialize\n");
    result = nn::act::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());

    // Initialize the Friend Presence library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:fp::Initialize\n");
    result = nn::fp::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    
    // Initialize the AC library
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Initialize\n");
    result = nn::ac::Initialize();
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Initialize:failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }

    // Connect to network using the default profile in the System Settings.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Connect:\n");
    result = nn::ac::Connect();
    if(result.IsFailure())
    {
        nn::ac::ErrorCode errorCode;
        nn::ac::GetLastErrorCode(&errorCode);
        PIA_SAMPLE_LOG("nn::ac::Connect() failed, errcode=%u\n", errorCode);

        PIA_SAMPLE_LOG("Inet_InitializeNetwork:ac::Connect:failed\n");
        PIA_TRACE_RESULT(result);
        return SampleFailedNetworkInterfaceUp();
    }

    // Set flag status to "connected."
    s_ACConnected = false;
    // Start the thread for checking the AC status.
    s_UpdateACStatusRun = true;
    s_UpdateACStatusThread.Start(UpdateACStatus, NULL, reinterpret_cast<void*>(s_UpdateACStatusThreadStack), sizeof(s_UpdateACStatusThreadStack), OS_PRIORITY_APP_DEFAULT);

    // Initialize the socket library.
    PIA_SAMPLE_LOG("Inet_InitializeNetwork:SOInit\n");
    if(SOInit() != 0)
    {
        PIA_SAMPLE_LOG("Inet_InitializeNetwork:SOInit failed, err=%d\n", SOLastError());
        return SampleFailedNetworkInterfaceUp();
    }

#elif NN_PIA_WIN
    // With Windows, do nothing because the network interface should be started automatically by the OS.
    
    // Initialize Winsock.
    WORD wVersionRequested = MAKEWORD(2, 2);    // Request the version of WinSock to use.
    WSADATA wsaData;                            // Winsock data structure.
    int nErrorStatus;

    nErrorStatus = WSAStartup(wVersionRequested, &wsaData);
    if (nErrorStatus != 0)
    {
        PIA_BASIC_REPORT("Winsock initialization failed\n");
        return SampleFailedNetworkInterfaceUp();
    }
    else
    {
        // Check whether it is same as requested version.
        if (LOBYTE( wsaData.wVersion ) == 2
        &&  HIBYTE( wsaData.wVersion ) == 2)
        {
            // Display version information.
            PIA_BASIC_REPORT("version %d.%d is required\n",
                LOBYTE(wsaData.wVersion),
                HIBYTE(wsaData.wVersion));
            PIA_BASIC_REPORT("latest version is %d.%d\n",
                LOBYTE(wsaData.wHighVersion),
                HIBYTE(wsaData.wHighVersion));
        }
    }
#endif

    // Prints the enabled network interface information.
    PrintIfConfig();

    return nn::ResultSuccess();
}

void Inet_FinalizeNetwork() {
#if NN_PIA_CTR
    nn::Result result;

    // Finalizing socket library
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork:socket::Finalize\n");
    result = nn::socket::Finalize();
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork:socket::Finalize:failed\n");
        PIA_TRACE_RESULT(result);
    }
    
    // Finalize event for receiving disconnection notifications.
    s_ACDisconnectedEvent.Finalize();
    
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Close\n");
    result = nn::ac::Close();
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Close:failed\n");
        PIA_TRACE_RESULT(result);
//n1589: If it returns here, the next AC initialization will fail.
    }

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Finalize\n");
    result = nn::ac::Finalize();
    if (result.IsFailure()) {
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Finalize:failed\n");
        PIA_TRACE_RESULT(result);
    }

// Need to call fs::Initialize, but no need to call fs::Finalize
#elif NN_PIA_CAFE
    nn::Result result;
    
    // If the ac::Connect function has succeeded, finalize the socket library, and join the thread for checking AC status.
    if (s_UpdateACStatusRun)
    {
        // Finalizing socket library
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork:SOFinish\n");
        int ret = SOFinish();
        if (ret != 0)
        {
            PIA_SAMPLE_LOG("Inet_FinalizeNetwork:SOFinish:failed, err=%d\n", ret);
        }

        // Terminate and join the thread for checking AC status.
        s_UpdateACStatusRun = false;
        s_UpdateACStatusThread.Join();
        s_UpdateACStatusThread.Finalize();
    }
    
    PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Close\n");
    result = nn::ac::Close();
    if(result.IsFailure())
    {
        nn::ac::ErrorCode errorCode;
        nn::ac::GetLastErrorCode(&errorCode);
        PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Close:failed, errcode=%u\n", errorCode);
        PIA_TRACE_RESULT(result);
    }

    PIA_SAMPLE_LOG("Inet_FinalizeNetwork:ac::Finalize\n");
    nn::ac::Finalize();
    nn::fp::Finalize();
    nn::act::Finalize();

#elif NN_PIA_WIN
    // Stop Winsock.
    WSACleanup();
    // With Windows, do nothing because the network interface is controlled by the OS.
#endif
}

bool Inet_IsConnected()
{
#if NN_PIA_CTR
    if (s_ACDisconnectedEvent.Wait(0)) {
        return false;
    } else {
        return true;
    }
#elif NN_PIA_CAFE
    return s_ACConnected;
#elif NN_PIA_WIN
    return true;
#endif
}
