/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Nex_Setup.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "SampleLib/SampleLib.h"
#include "Shared_Nex_Setup.h"
//#include "Shared_Nex_ErrorHandling.h"

#include "Shared_Inet_Matchmake.h"

#if NN_PIA_CAFE
#include <cafe/mem.h>
#endif

namespace {
// Multiple calls for finalization are OK.
    bool s_Initialized = false;

    // Main memory size
    size_t s_MemorySize = 0;

    // Emergency memory size
    size_t s_EmergencyMemorySize = 0;

    // Main memory
    void* s_pMemory         = NULL;

    // Emergency memory used when the heap is depleted
    void* s_pEmergencyMemory = NULL;

    // Memory usage
    size_t s_MemUsage    = 0;
    size_t s_MemMaxUsage = 0;

#if NN_PIA_CTR
    // Heap for NEX
    nn::fnd::ThreadSafeExpHeap s_NexHeap;
#elif NN_PIA_CAFE
    MEMHeapHandle s_NexHeap;
#endif

#if NN_PIA_NEX_3
    //P2P signature key. Must be changed for each application.
    const nn::nex::qByte SIGNATURE_KEY[] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09}; 
#endif
    // Pointer to shared session key obtained from server when the matchmaking session was established.
    // Only allocate instances of qVector and other NEX variables between the initialization and finalization of the NEX heap, because memory for these variables is allocated from the NEX heap.
    nn::nex::qVector<nn::nex::qByte>* s_pMatchmakeSessionKey = NULL;

#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
    // Multi-unit heap
    typedef nn::nex::ThreadUnsafeMultiUnitHeap TMultiUnitHeap;
#endif
    bool s_UseMutliUnitHeap = false;

//-----------------------------------------------------------------------------------------
// Declaration of Internal Functions
//-----------------------------------------------------------------------------------------
    void InitializeRuntimeMemory(size_t memorySize, size_t emergencyMemorySize);

    void FinalizeRuntimeMemory();

#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
    void InitializeMultiUnitHeap(size_t memorySize, size_t emergencyMemorySize, bool useNetZ);

    void FinalizeMultiUnitHeap();
#endif

    void* InnerAlloc(size_t size);
    void  InnerFree(void* ptr);

    void* AllocFunction(size_t size);
    void FreeFunction(void *pointer);

    void AcquireReservedMemory();
    void ReleaseReservedMemory();
}

//-----------------------------------------------------------------------------------------
// Implementation of External Functions
//-----------------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------------
// NEX initialization
//-----------------------------------------------------------------------------------------
nn::Result Nex_InitializeNex(
    size_t memorySize, 
    size_t emergencyMemorySize, 
    nn::nex::Core::ThreadMode threadMode,
    bool useNetZ,
    bool useMultiUnitHeap
) {
    PIA_SAMPLE_LOG("Nex_InitializeNex\n");

    // Initialization of NEX heap
    s_UseMutliUnitHeap = useMultiUnitHeap;
    if(s_UseMutliUnitHeap) {
#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
        PIA_SAMPLE_LOG("Nex_InitializeNex:InitializeMultiUnitHeap\n");
        InitializeMultiUnitHeap(memorySize, emergencyMemorySize, useNetZ);
#else
        (void)useNetZ;
        
        PIA_SAMPLE_LOG("Nex_InitializeNex:InitializeRuntimeMemory\n");
        InitializeRuntimeMemory(memorySize, emergencyMemorySize);
#endif
    }
    else {
        PIA_SAMPLE_LOG("Nex_InitializeNex:InitializeRuntimeMemory\n");
        InitializeRuntimeMemory(memorySize, emergencyMemorySize);
    }

    // Initialization of global instances, etc.
    nn::nex::GlobalVariables::AcquireInstance();

    // Initialize error handling.
//    Nex_InitializeErrorHandling();

    // Sets the network topology
//    nn::nex::NetZ::SetNetworkTopology(nn::nex::NetworkTopology::Peer2Peer);

    // Thread settings
    nn::nex::Core::SetThreadMode(threadMode);

#if NN_PIA_NEX_3
    // Set P2P's signature key.
    nn::nex::Network::GetInstance()->SetP2PDataPacketDefaultSignatureKey(SIGNATURE_KEY, sizeof(SIGNATURE_KEY));
#endif

    // Initialize matchmaking variables.
    if (s_pMatchmakeSessionKey == NULL) {
        s_pMatchmakeSessionKey = qNew nn::nex::qVector<nn::nex::qByte>();
    }

    s_Initialized = true;
    return nn::ResultSuccess();
}

//-----------------------------------------------------------------------------------------
// NEX finalization
//-----------------------------------------------------------------------------------------
void Nex_FinalizeNex() {
    if (!s_Initialized) {
        PIA_SAMPLE_LOG("Nex_FinalizeNex:already finalized\n");
        return;
    }

    PIA_SAMPLE_LOG("Nex_FinalizeNex:\n");

    // Finalize matchmaking variables.
    if (s_pMatchmakeSessionKey != NULL) {
        s_pMatchmakeSessionKey->clear();
        qDelete s_pMatchmakeSessionKey;
        s_pMatchmakeSessionKey = NULL;
    }

    // Finalization of global instances.
    nn::nex::GlobalVariables::ReleaseInstance();

    // Finalize error handling.
//    Nex_FinalizeErrorHandling();

    // Finalization of NEX heap
    if(s_UseMutliUnitHeap) {
#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
        FinalizeMultiUnitHeap();
#else
        FinalizeRuntimeMemory();
#endif
    }
    else {
        FinalizeRuntimeMemory();
    }

    s_Initialized = false;
}

//-----------------------------------------------------------------------------------------
// Records the peak values of multi-unit heaps.
//-----------------------------------------------------------------------------------------
void Nex_EnableMultiUnitHeapPeakRecording(void) {
#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
    TMultiUnitHeap::EnablePeakRecording();
#endif
}

void Nex_DisableMultiUnitHeapPeakRecording(void) {
#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
    TMultiUnitHeap::DisablePeakRecording();
#endif
}

size_t Nex_GetMultiUnitHeapPeakMemorySize(void) {
#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
    bool ret;

    size_t unitSize[TMultiUnitHeap::LEVEL] = {};
    ret = TMultiUnitHeap::GetUnitSizes(unitSize);
    if(!ret)
    {
        return 0;
    }

    size_t peakCount[TMultiUnitHeap::LEVEL] = {};
    ret = TMultiUnitHeap::GetPeakCounts(peakCount);
    if(!ret)
    {
        return 0;
    }

    size_t peakTotal = 0;
    for(int i=0; i<TMultiUnitHeap::LEVEL; ++i)
    {
        peakTotal += (unitSize[i] * peakCount[i]);
    }

    return peakTotal;
#else
    return 0;
#endif
}


//-----------------------------------------------------------------------------------------
// Outputs protocol call context error to the log
//-----------------------------------------------------------------------------------------
void Nex_ReportProtocolCallContext(nn::nex::ProtocolCallContext& context, const char* pHeaderText) {
    // Suppresses the warnings for unused parameters
    (void)context;

    if (pHeaderText)
    {
        PIA_SAMPLE_LOG(pHeaderText);
    }

    PIA_SAMPLE_LOG("Nex_ReportProtocolCallContext:(Facility:%d Code:0x%X NErr:%d)\n", 
        context.GetOutcome().GetFacility(), 
        context.GetOutcome().GetReturnCode(),
        nn::nex::ErrorCodeConverter::ConvertToNetworkErrorCode(context.GetOutcome()));
}


//-----------------------------------------------------------------------------------------
// Function to access shared session key from Inet_AutoMatchmake, etc.
//-----------------------------------------------------------------------------------------
nn::nex::qVector<nn::nex::qByte>* GetMatchmakeSessionKeyPtr() {
    return s_pMatchmakeSessionKey;
}


//-----------------------------------------------------------------------------------------
// Implementation of Internal Functions
//-----------------------------------------------------------------------------------------
namespace {
//-----------------------------------------------------------------------------------------
// Initialization of NEX heap (ExpHeap)
//-----------------------------------------------------------------------------------------
void InitializeRuntimeMemory(size_t memorySize, size_t emergencyMemorySize) {
    s_MemorySize = memorySize;
    s_EmergencyMemorySize = emergencyMemorySize;

    size_t total = memorySize + emergencyMemorySize;
    PIA_SAMPLE_LOG("InitializeRuntimeMemory:MemorySize:%d, ReservedMemorySize:%d, Total:%d\n", s_MemorySize, s_EmergencyMemorySize, total);

#if NN_PIA_CTR
    // Allocate work memory
    s_pMemory = new u8[total];
    // Initialize heap
    s_NexHeap.Initialize(NN_ANY_TO_UPTR(s_pMemory), total);
#elif NN_PIA_WIN
    // Allocate work memory
    s_pMemory = new u8[total];
#elif NN_PIA_CAFE
    // Allocate work memory from the default heap (aligned to 8 bytes)
    s_pMemory = MEMAllocFromDefaultHeapEx(total, 8);
    // Initialize heap
    s_NexHeap = MEMCreateExpHeapEx(s_pMemory, total, MEM_HEAP_OPT_0_CLEAR | MEM_HEAP_OPT_THREAD_SAFE);
#endif

    // Registers memory function
    nn::nex::MemoryManager::SetBasicMemoryFunctions(AllocFunction, FreeFunction);

    if (s_EmergencyMemorySize != 0) {
        // Allocates reserved memory for emergency
        AcquireReservedMemory();
    }
}

//-----------------------------------------------------------------------------------------
// Finalization of NEX heap (ExpHeap)
//-----------------------------------------------------------------------------------------
void FinalizeRuntimeMemory() {
    if (s_EmergencyMemorySize != 0) {
        // Release if there is reserved memory for emergency left over
        ReleaseReservedMemory();
    }

    PIA_SAMPLE_PRINT(
        "FinalizeRuntimeMemory:Nex memory usage: Now memory usage:%d, Max memory usage:%d / %d\n", 
        s_MemUsage, 
        s_MemMaxUsage, 
        s_MemorySize + s_EmergencyMemorySize
    );

#if NN_PIA_CTR
    // Memory is leaking
    if (s_MemUsage > 0) {
        // Displays the memory that is leaking
        s_NexHeap.Dump();
        // Disables memory that is leaking
        s_NexHeap.Invalidate();
    }
    // Destroys heap.
    s_NexHeap.Finalize();
    // Destroys work memory
    delete[] s_pMemory;
#elif NN_PIA_WIN
    // Destroys work memory
    delete[] s_pMemory;
#elif NN_PIA_CAFE
    // Memory is leaking
    if (s_MemUsage > 0) {
        // Dumps the heap information.
        MEMDumpHeap(s_NexHeap);
    }
    // Destroys the heap.
    MEMDestroyExpHeap(s_NexHeap);
    // Deallocate work memory
    MEMFreeToDefaultHeap(s_pMemory);
#endif


    s_MemUsage    = 0;
    s_MemMaxUsage = 0;
}

#if (NN_PIA_NEX_3==0) && (NN_PIA_EXPERIMENT_UNUSE_NEX_MULTIUNITHEAP==0)
//-----------------------------------------------------------------------------------------
// Initializes NEX heap (MultiUnitHeap)
//-----------------------------------------------------------------------------------------
void InitializeMultiUnitHeap(size_t memorySize, size_t emergencyMemorySize, bool useNetZ) {
    s_MemorySize = memorySize;
    s_EmergencyMemorySize = emergencyMemorySize;

    size_t total = memorySize + emergencyMemorySize;
    PIA_SAMPLE_LOG("InitializeMultiUnitHeap:MemorySize:%d, ReservedMemorySize:%d, Total:%d\n", s_MemorySize, s_EmergencyMemorySize, total);

    s_pMemory = new u8[total];

    const int LEVEL = 5;
    size_t unitSize[LEVEL];
    size_t unitObjs[LEVEL];
    if(useNetZ) {
        unitSize[0] =      100;  unitObjs[0] = 4000;
        unitSize[1] =      300;  unitObjs[1] = 2000;
        unitSize[2] =     1200;  unitObjs[2] =  800;
        unitSize[3] =     1800;  unitObjs[3] =  100;
        unitSize[4] =   8*1024;  unitObjs[4] =   40;
    }
    else {
        unitSize[0] =      256;  unitObjs[0] = 1024;
        unitSize[1] =     2048;  unitObjs[1] =  256;
        unitSize[2] =   4*1024;  unitObjs[2] =   32;
        unitSize[3] =   8*1024;  unitObjs[3] =    4;
        unitSize[4] =  24*1024;  unitObjs[4] =    1;
    }

    bool ret = TMultiUnitHeap::Initialize(s_pMemory, total, unitSize, unitObjs, LEVEL, emergencyMemorySize);
    PIA_SAMPLE_ASSERTMSG(ret, "InitializeMultiUnitHeap() failed.");
    NN_PIA_DUMMY_PARAM(ret);
}

//-----------------------------------------------------------------------------------------
// Finalizes NEX heap (MultiUnitHeap)
//-----------------------------------------------------------------------------------------
void FinalizeMultiUnitHeap(void) {
    size_t memSize = Nex_GetMultiUnitHeapPeakMemorySize();
    if(memSize > 0) {
        PIA_SAMPLE_LOG("MultiUnitHeap peak memory usage: %d bytes (%d KB)\n", memSize, memSize/1024);
    }
    else {
        PIA_SAMPLE_LOG("Perhaps MultiUnitHeap memory peak recording feature is disabled.\n");
    }

    TMultiUnitHeap::Finalize();

    delete[] s_pMemory;
}
#endif

//-----------------------------------------------------------------------------------------
// Application-defined memory allocation function
//-----------------------------------------------------------------------------------------
void* AllocFunction(size_t size) {
    void* p = InnerAlloc(size);

    if (p == NULL) {
        PIA_SAMPLE_LOG("AllocFunction: No enough memory. Reserved memory will be used. Please shutdown NEX library.\n");
        ReleaseReservedMemory();

        p = InnerAlloc(size);
        PIA_SAMPLE_ASSERT(p != NULL);
        if (p == NULL) {
            PIA_PANIC("Reserved memory is not enough");
        }
    }

    return p;
}

//-----------------------------------------------------------------------------------------
// Application-defined memory deallocation function
//-----------------------------------------------------------------------------------------
void FreeFunction(void *pointer) {
    InnerFree(pointer);
}

//-----------------------------------------------------------------------------------------
// Allocates memory for emergency
//-----------------------------------------------------------------------------------------
void AcquireReservedMemory() {
    if (s_pEmergencyMemory == NULL) {
        PIA_SAMPLE_LOG("AcquireReservedMemory\n");

        size_t nowUsage    = s_MemUsage;
        size_t nowMaxUsage = s_MemMaxUsage;

        s_pEmergencyMemory = AllocFunction(s_EmergencyMemorySize);

        s_MemUsage    = nowUsage;
        s_MemMaxUsage = nowMaxUsage;
    }
}

//-----------------------------------------------------------------------------------------
//Deallocates memory for emergency
//-----------------------------------------------------------------------------------------
void ReleaseReservedMemory() {
    if (s_pEmergencyMemory) {
        PIA_SAMPLE_LOG("ReleaseReservedMemory\n");

        size_t nowUsage    = s_MemUsage;
        size_t nowMaxUsage = s_MemMaxUsage;

        FreeFunction(s_pEmergencyMemory);

        s_pEmergencyMemory = NULL;

        s_MemUsage    = nowUsage;
        s_MemMaxUsage = nowMaxUsage;
    }
}

//-----------------------------------------------------------------------------------------
// Internal memory function called from AllocFunction
//-----------------------------------------------------------------------------------------
void* InnerAlloc(size_t size) {
#if NN_PIA_CTR
    void* pBuffer = s_NexHeap.Allocate(size);
#elif NN_PIA_WIN
    void* pBuffer = malloc(size);
#elif NN_PIA_CAFE
    void* pBuffer = MEMAllocFromExpHeapEx(s_NexHeap, size, 8);
#endif
    if (pBuffer == NULL) {
        return NULL;
    }

#if NN_PIA_CTR
    s_MemUsage += s_NexHeap.GetSizeOf(pBuffer);
    if (s_MemUsage > s_MemMaxUsage) {
        s_MemMaxUsage = s_MemUsage;
    }
#elif NN_PIA_CAFE
    s_MemUsage += MEMGetSizeForMBlockExpHeap(pBuffer);
    if (s_MemUsage > s_MemMaxUsage) {
        s_MemMaxUsage = s_MemUsage;
    }
#endif

    return pBuffer;
}

//-----------------------------------------------------------------------------------------
// Internal memory function called from FreeFunction
//-----------------------------------------------------------------------------------------
void InnerFree(void* ptr) {
#if NN_PIA_CTR
    s_MemUsage -= s_NexHeap.GetSizeOf(ptr);
    s_NexHeap.Free(ptr);
#elif NN_PIA_CAFE
    s_MemUsage -= MEMGetSizeForMBlockExpHeap(ptr);
    MEMFreeToExpHeap(s_NexHeap, ptr);
#elif NN_PIA_WIN
    free(ptr);
#endif
}

} // Namespace
