﻿/*---------------------------------------------------------------------------*
  Project:  Pia
  File:     Shared_Session_Setup.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include "SampleLib/SampleLib.h"
#include "Shared_Session_Setup.h"
#include "Shared_Nex_Common.h"
#if NN_PIA_CAFE
#include <cafe.h>
#include <nn/act.h>
#include <nn/fp.h>
#elif NN_PIA_CTR
#include <nn/friends.h>
#endif

nn::Result Session_InitializePiaSession( nn::pia::transport::NetworkFactory* pFactory ) {
    nn::pia::session::Session::Setting setting = {
        pFactory,
        // Configure the network topology
        nn::pia::session::NetworkTopology_FullMesh,
        // Configure the host migration mode
        nn::pia::session::HostMigrationMode_None
    };
    return Session_InitializePiaSession( setting );
}

nn::Result Session_InitializePiaSession( const nn::pia::session::Session::Setting& setting ) {
    PIA_SAMPLE_LOG("Session_InitializePiaSession\n");
    nn::Result result;

    result = nn::pia::session::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::Session::CreateInstance(setting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);   

    // Print out the amount of consumed memory
    PIA_SAMPLE_PRINT("Session_InitializePiaSession:Pia current memory usage: %d\n", nn::pia::common::GetMemoryUsage());
    return result;
}

void Session_FinalizePiaSession() {
    PIA_SAMPLE_LOG("Session_FinalizePiaSession\n");

// With Pia, Finalize and DestroyInstance always succeed
    if (nn::pia::session::Session::GetInstance() != NULL) {
        nn::pia::session::Session::DestroyInstance();
    }
    nn::pia::session::Finalize();
}


void GetLocalPlayerName(nn::pia::transport::Station::PlayerName* pPlayerName) {

    nn::Result result;

#if NN_PIA_CAFE
    //Use the Mii character's name.
    char16 miiName[ACT_MII_NAME_SIZE];
    std::memset(miiName, 0, sizeof(miiName));
    
    result = nn::act::GetMiiName(miiName);
    if(result.IsSuccess())
    {
        for (int i=0; i<ACT_MII_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (miiName[i] != 0)
            {
                pPlayerName->data[i] = miiName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    pPlayerName->language = nn::fp::LANGUAGE_JAPANESE;
#elif NN_PIA_CTR
    //Use the Mii character's name.
    char16 screenName[nn::friends::SCREEN_NAME_SIZE];
    result = nn::friends::Initialize();
    PIA_SAMPLE_ASSERT(result.IsSuccess());
    result = nn::friends::GetMyScreenName(screenName);
    if(result.IsSuccess() && screenName[0] != 0) // Also exclude cases where a Mii character is not set.
    {
        for (int i=0; i<nn::friends::SCREEN_NAME_SIZE; i++)
        {
            pPlayerName->length = i;
            if (screenName[i] != 0)
            {
                pPlayerName->data[i] = screenName[i];
            }
            else
            {
                break;
            }
        }
    }
    else
    {
        std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
        pPlayerName->length = 9;
    }
    pPlayerName->language = 0;
#else
    std::memcpy(pPlayerName->data, reinterpret_cast<const s16*>(_T("DummyName")), 18);
    pPlayerName->length = 9;
    pPlayerName->language = 0;
#endif

}

nn::Result Session_StartupSession(
    const nn::pia::common::SignatureSetting& signatureSetting,
    const nn::pia::transport::Station::IdentificationToken* pToken
) {
    PIA_SAMPLE_LOG("Session_StartupSession\n");

    nn::Result result;
    nn::pia::transport::Station::PlayerName playerName;
    
    GetLocalPlayerName(&playerName);

    result = nn::pia::session::Session::GetInstance()->Startup(
        &playerName,                // The player name indicating the local host
        signatureSetting,           // Configure packet with signature feature.
        pToken                      // The identification token indicating the local host
    );
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    return result;
}

nn::Result Session_StartupSession(
    u32 maxSilenceTime,
    u32 keepAliveSendingInterval,
    const nn::pia::common::SignatureSetting& signatureSetting,
    const nn::pia::transport::Station::IdentificationToken* pToken
    ) {

    nn::Result result;
    nn::pia::transport::Station::PlayerName playerName;
    
    GetLocalPlayerName(&playerName);

    result = nn::pia::session::Session::GetInstance()->Startup(
        maxSilenceTime,
        keepAliveSendingInterval,
        &playerName,                // The player name indicating the local host
        signatureSetting,           // Configure packet with signature feature.
        pToken                      // The identification token indicating the local host
    );
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);
    return result;
}

void Session_CleanupSession()
{
    PIA_SAMPLE_LOG("Session_CleanupSession\n");
    nn::pia::session::Session::GetInstance()->Cleanup();
#if NN_PIA_CTR
    nn::friends::Finalize();
#endif
}

