/*---------------------------------------------------------------------------*
  Project:  Voice Chat Sample.
  File:     net.cpp

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*
    Include
 *---------------------------------------------------------------------------*/
#include "net.h"
#include "framework.h"
#include "Shared/Shared_Inet_Login.h"
#include "Shared/Shared_Inet_Matchmake.h"
#include "Shared/Shared_Inet_NetSetup.h"
#include "Shared/Shared_Nex_Setup.h"
#include "Shared/Shared_Inet_Setup.h"
#include "Shared/Shared_Session_Setup.h"

/*---------------------------------------------------------------------------*
    Define
 *---------------------------------------------------------------------------*/
#define STACK_SIZE  4096*20
#ifdef VCT_PLATFORM_CTR
#define GAME_ID 0x1F906E00
#else
#define GAME_ID 0x1F906E00
#endif
#define ACCESS_KEY _T("3c273382")


CriticalSection s_NodeStatusLock;
#ifdef VCT_PLATFORM_CTR
#else
u8 s_ThreadStack[STACK_SIZE];
#endif
struct SampleNodeStatus
{
    nn::pia::StationId           stationId;
};

const u32 MEMORY_BLOCK_SIZE_NEX             = 1024 * 1024;      //!< Specifies the memory size provided to NEX.
const u32 MEMORY_BLOCK_SIZE_NEX_EMERGENCY   = 50 * 1024;        //!< Specifies the memory size provided to NEX for emergencies.
const nn::nex::Core::ThreadMode SAMPLE_NEX_THREAD_MODE = nn::nex::Core::ThreadModeUnsafeTransportBuffer;

/*---------------------------------------------------------------------------*
    Anonymous Namespace
 *---------------------------------------------------------------------------*/
namespace {
    SampleNodeStatus   s_NodeStatus[ NET_MAX_CONNECT ];
    u32                s_ConnectionNum;
    NetReceiveCallback receiveCallback;

    // Related to the Pia protocol.
    u32 s_UnreliableProtocolHandle = 0;
    u32 s_ReliableProtocolHandle = 0;
}

/*---------------------------------------------------------------------------*
    Prototype
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
void NetThread( NetworkInfra* net );
#else
int NetThread( int,void* );
#endif

/*---------------------------------------------------------------------------*
    Function
 *---------------------------------------------------------------------------*/

u32 ConvertID_Pia2VCT(u32 id)
{
    return(id+1);
}
u32 ConvertID_VCT2Pia(u32 id)
{
    return(id-1);
}

void sessionCallback(nn::pia::session::Session::EventType eventType, nn::pia::StationId id)
{
    SAMPLE_PRINT("%s is call(%d, %d)\n", __func__, eventType, id);

    s_NodeStatusLock.Lock();

    switch( eventType )
    {
        case nn::pia::session::Session::EVENT_JOIN:
            s_NodeStatus[id].stationId = id; //Only shows whether it is valid.
            s_ConnectionNum++;
            break;
        case nn::pia::session::Session::EVENT_LEAVE:
            s_NodeStatus[id].stationId = nn::pia::STATION_ID_INVALID; //Only shows whether it is valid.
            s_ConnectionNum--;
            break;
    }

    s_NodeStatusLock.Unlock();
}


bool NetworkInfra::Initialize( void )
{
    s_NodeStatusLock.Initialize();
    m_PiaCommon = new PiaNetworkUtil;
    return m_PiaCommon->InitializeCommon();
}

void NetworkInfra::Finalize( void )
{
    if(m_PiaCommon != NULL){
        m_PiaCommon->FinalizeCommon();
        delete m_PiaCommon;
    }
    s_NodeStatusLock.Finalize();
}

/*---------------------------------------------------------------------------*
  Name        : Start
  Description : Start Infra.
  Arguments   : -
  Returns     : -
 *---------------------------------------------------------------------------*/
void NetworkInfra::Start( NetConfigStructure *config )
{
    // Copy Config...
    m_Config.eventCallbackFunction = config->eventCallbackFunction;
    receiveCallback               = config->receiveCallbackFunction;
    m_Config.gameMode             = config->gameMode;
    m_Config.maxConnect           = config->maxConnect;
    m_Config.netMatchMode         = config->netMatchMode;

    // Initialize Critical Section
    m_SendCC.Initialize();

    // Infra thread start.
    m_NetThreadEndFlg = 0;
#ifdef VCT_PLATFORM_CTR
    m_NetThread.StartUsingAutoStack( NetThread, this, STACK_SIZE, NN_OS_DEFAULT_THREAD_PRIORITY - 1 );
#else
    if(OSGetCoreCount() == 3)
    {
        u32 running_core = OSGetCoreId();
        u32 main_core = OSGetMainCoreId();
        u32 sub_core = 2;
        if(main_core == 0){
            sub_core = 2;
        }
        else if(main_core == 2)
        {
            sub_core = 0;
        }
        SAMPLE_PRINT("Net:Run=%d Main=%d Sub=%d\n",running_core,main_core,sub_core);
        m_NetThread = OSGetDefaultThread( sub_core );
        if(OSRunThread( m_NetThread, NetThread, 0, (void *)this) == FALSE)
        {
            OSHalt("Can't run net thead\n");
        }
    }
    else
    {
        OSHalt("Can't run net thead\n");
    }
#endif
}


/*---------------------------------------------------------------------------*
  Name        : Stop
  Description : Stop Infra.
  Arguments   : -
  Returns     : -
 *---------------------------------------------------------------------------*/
void NetworkInfra::Stop( void )
{
    // Infra thread stop.
    m_NetThreadEndFlg = 1;
#ifdef VCT_PLATFORM_CTR
    m_NetThread.Join();
    m_NetThread.Finalize();
#else
    while( OSIsThreadTerminated(m_NetThread) != TRUE ){
        OSSleepMilliseconds(16);
    }
    //OSJoinThread(m_NetThread,NULL);
    SAMPLE_PRINT("netInfra->Stop()\n");
#endif
    // Finalize Critical Section
    m_SendCC.Finalize();
}


/*---------------------------------------------------------------------------*
  Name        : NetThreadImpl
  Description : Infra main thread.
  Arguments   : -
  Returns     : -
 *---------------------------------------------------------------------------*/
void NetworkInfra::NetThreadImpl( void )
{
    bool bret;
    s32 sret;
    nn::Result result;
    nn::pia::StationId srcId;
    nn::pia::StationId myStationId;
#ifdef VCT_PLATFORM_CTR
    nn::socket::InAddr addr1,addr2;
#else
    nn::ac::IpAddress addr;
#endif
    bit8 recbuf[nn::pia::transport::UnreliableProtocol::MAX_DATA_SIZE];  // Allocate an array with the maximum size that could arrive.

    // Variable Initialize...
    s_ConnectionNum     = 0;
    m_JoinedGatheringID = INVALID_GATHERINGID;
    m_ReqSendDataIndxR  = 0;
    m_ReqSendDataIndxW  = 0;
    std::memset( m_SendDataInfo, 0x0, sizeof(NetSendDataStructure) * NET_SEND_DATA_MAX );
    m_Info.connectionNum = 0;
    m_Info.ownerPid = 0;
    m_Info.myPid = 0;
    m_Info.myCid = 0;
    std::memset( &m_Info.CIDList,        0x0, sizeof(u32)*NET_MAX_CONNECT );
    for(sret = 0 ; sret < NET_MAX_CONNECT ; sret++){
        m_Info.CIDList[sret] = nn::pia::STATION_ID_INVALID;
    }
    m_SessionStartFlg   = false;
    m_NetErrorFlg       = false;

    // NEX Start...
    bret = m_PiaCommon->InitializeInet();
    if (bret == false )
    {
        m_NetErrorFlg = true;
        goto NetThreadImpl_Term0_;
    }

#ifdef VCT_PLATFORM_CTR
    sret = nn::socket::GetPrimaryAddress(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2)); 
    PIA_ASSERT(sret == 0);
    m_Info.hostAddr.SetInetAddress(nn::pia::common::InetAddress(nn::pia::common::ByteOrder::Swap32(addr1.addr), 0));
    m_Info.netMask.SetInetAddress(nn::pia::common::InetAddress(nn::pia::common::ByteOrder::Swap32(addr2.addr), 0));

    sret = nn::socket::GetResolver(reinterpret_cast<u8*>(&addr1), reinterpret_cast<u8*>(&addr2));
    PIA_ASSERT(sret == 0);
    m_Info.dns1.SetInetAddress(nn::pia::common::InetAddress(nn::pia::common::ByteOrder::Swap32(addr1.addr), 0));
    m_Info.dns2.SetInetAddress(nn::pia::common::InetAddress(nn::pia::common::ByteOrder::Swap32(addr2.addr), 0));

    sret = nn::socket::GetDefaultGateway(reinterpret_cast<u8*>(&addr1));
    PIA_ASSERT(sret == 0);
    m_Info.gateway.SetInetAddress(nn::pia::common::InetAddress(nn::pia::common::ByteOrder::Swap32(addr1.addr), 0));

    sret = nn::socket::GetMacAddress (reinterpret_cast<u8*>(m_Info.mac));
    PIA_ASSERT(sret == 0);
#else
    result = nn::ac::GetAssignedAddress(&addr);
    PIA_ASSERT(result.IsSuccess());
    m_Info.hostAddr.SetInetAddress(nn::pia::common::InetAddress(addr, 0));

    result = nn::ac::GetAssignedSubnet(&addr);
    PIA_ASSERT(result.IsSuccess());
    m_Info.netMask.SetInetAddress(nn::pia::common::InetAddress(addr, 0));

    result = nn::ac::GetAssignedPreferedDns(&addr);
    PIA_ASSERT(result.IsSuccess());
    m_Info.dns1.SetInetAddress(nn::pia::common::InetAddress(addr, 0));

    result = nn::ac::GetAssignedAlternativeDns(&addr);
    PIA_ASSERT(result.IsSuccess());
    m_Info.dns2.SetInetAddress(nn::pia::common::InetAddress(addr, 0));

    result = nn::ac::GetAssignedGateway(&addr);
    PIA_ASSERT(result.IsSuccess());
    m_Info.gateway.SetInetAddress(nn::pia::common::InetAddress(addr, 0));
    //Cafe does not allow acquisition of the MAC address.
#endif

    // Matchmaking.
    bret = m_PiaCommon->AutoMatch(m_Config.maxConnect,(m_Config.netMatchMode == NET_MATCH_ANYBODY)?nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY : nn::nex::MATCHMAKE_SYSTEM_TYPE_FRIEND);
    if ( bret == false )
    {
        m_NetErrorFlg = true;
        goto NetThreadImpl_Term0_;
    }

    // Session.
    bret = m_PiaCommon->Session();
    if ( bret == false )
    {
        m_NetErrorFlg = true;
        goto NetThreadImpl_Term1_;
    }

    // Get the Connection ID.
    myStationId = nn::pia::session::Session::GetInstance()->GetLocalStationId();
    SAMPLE_PRINT(" my connection id (0x%x)\n", myStationId);
    m_Info.myCid = myStationId;
    //s_NodeStatus[myStationId].stationId = myStationId;
    //s_ConnectionNum++;
    m_Info.ownerPid = m_PiaCommon->GetOwnerPID();
    m_Info.myPid = m_PiaCommon->GetMyPID();

    // Event Callback... Connect...
    if ( m_Config.eventCallbackFunction != NULL )
    {
        m_Config.eventCallbackFunction( NET_EVENT_CONNECT, &m_Info );
    }
    // The session start flag is enabled.
    m_SessionStartFlg = true;

    // Main Loop
    while ( true )
    {
        //Disconnection detected.
        //Status check for the network connection.
        result = nn::pia::transport::Transport::GetInstance()->GetConnectionStatus();
        if ( result.IsFailure() )
        {
            //If the network connection is abnormal:
            if ( result == nn::pia::ResultConnectionIsLost() )
            {
                // Returns when the network connection is interrupted because the wireless switch is turned off or communication fails.
                SAMPLE_PRINT("Connection Lost when scene Communication.\n");
            }
            else
            {
                //Unexpected state.
                SAMPLE_PRINT("Transport::GetConnectionStatus is unknown status when scene Communication.\n");
                PIA_TRACE_RESULT(result);
                PIA_SAMPLE_ASSERT(false);
            }
            break;
        }
        //Session status check.
        if( nn::pia::session::Session::GetInstance()->GetState() == nn::pia::session::Session::STATE_DISCONNECTED )
        {
            SAMPLE_PRINT("SessionState became DISCONNECTED.\n");
            break; //In this case, exiting is forced.
        }
        
        // End Check.
        if (m_NetThreadEndFlg == 1)
        {
            break;
        }

        // Connection Check.
        bool connection_change_flg = false;
        for ( int i = 0; i < NET_MAX_CONNECT; i++)
        {
            if ( m_Info.CIDList[i] != s_NodeStatus[i].stationId )
            {
                connection_change_flg = true;
                m_Info.CIDList[i] = s_NodeStatus[i].stationId;
            }
        }
        if ( m_Info.connectionNum != s_ConnectionNum )
        {
            connection_change_flg = true;
            m_Info.connectionNum = s_ConnectionNum;
            SAMPLE_PRINT("ConnectionNum : %d\n", s_ConnectionNum);
        }

        if ( connection_change_flg )
        {
            // Event Callback... Connection Change...
            if ( m_Config.eventCallbackFunction != NULL )
            {
                m_Config.eventCallbackFunction( NET_EVENT_CHANGE_CONNECTION, &m_Info );
            }
        }

        // Execute Dispatch (first time: for receive processing).
        size_t recvSize;
        nn::pia::common::Scheduler::GetInstance()->Dispatch();
        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Receive(&srcId,recbuf,&recvSize,sizeof(recbuf));
        if(result.IsSuccess())
        {
            receiveCallback(static_cast<s32>(srcId),(void*)recbuf,static_cast<u32>(recvSize));
        }

        result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_ReliableProtocolHandle)->Receive(&srcId,recbuf,&recvSize,sizeof(recbuf));
        if(result.IsSuccess())
        {
            receiveCallback(static_cast<s32>(srcId),(void*)recbuf,static_cast<u32>(recvSize));
        }

        if ( m_ReqSendDataIndxW != m_ReqSendDataIndxR )
        {
            u32  *buf       = (u32 *)m_SendDataInfo[m_ReqSendDataIndxR].send_buff;
            u32  destCID    = m_SendDataInfo[m_ReqSendDataIndxR].destCid;
            u32  size       = m_SendDataInfo[m_ReqSendDataIndxR].dataSize;
            bool isReliable = m_SendDataInfo[m_ReqSendDataIndxR].reliable;

            if ( isReliable )
            {
                 nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_ReliableProtocolHandle)->Send( s_NodeStatus[destCID].stationId, (u8 *)buf, size );
            }
            else
            {
                 nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Send( s_NodeStatus[destCID].stationId, (u8 *)buf, size );
            }

            // Update...
            m_ReqSendDataIndxR = ( m_ReqSendDataIndxR + 1 ) % NET_SEND_DATA_MAX;
        }

        // To maintain communication with the NEX game server, you must call nn::nex::Scheduler::Dispatch.
        // In this demo, it is called in every frame, but you may reduce the call frequency to reduce processing.
        nn::nex::Scheduler::GetInstance()->Dispatch( 3 );

        // Execute Dispatch (second time: for send processing).
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

#ifdef VCT_PLATFORM_CTR
        nn::os::Thread::Sleep( nn::fnd::TimeSpan::FromMilliSeconds( 8 ) );
#else
        OSSleepMilliseconds( 8 );
#endif
    }

NetThreadImpl_Term1_:
    // The session processing is finished, so Cleanup runs.
    nn::pia::session::Session::GetInstance()->Cleanup();
    if (nn::pia::session::Session::GetInstance() != NULL) {
        nn::pia::session::Session::DestroyInstance();
    }
    nn::pia::session::Finalize();
    // Finish Cleanup of NexFacade before before finalizing transport.
    if( nn::pia::inet::NexFacade::GetInstance() != NULL )
    {
        nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
        nn::pia::inet::NexFacade::GetInstance()->Cleanup();
    }

    if (nn::pia::transport::Transport::GetInstance() != NULL) {
        if( nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle) != NULL )
        {
            nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Finalize();
            nn::pia::transport::Transport::GetInstance()->DestroyProtocol(s_UnreliableProtocolHandle);
        }
        if( nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_ReliableProtocolHandle) != NULL )
        {
            nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_ReliableProtocolHandle)->Finalize();
            nn::pia::transport::Transport::GetInstance()->DestroyProtocol(s_ReliableProtocolHandle);
        }
    }
    if (nn::pia::transport::Transport::GetInstance() != NULL) {
        nn::pia::transport::Transport::DestroyInstance();
    }
    nn::pia::transport::Finalize();

NetThreadImpl_Term0_:
    // NEX stop...
    m_PiaCommon->FinalizeInet();
    // The session start flag is disabled.
    m_SessionStartFlg = false;
    SAMPLE_PRINT( "End net loop.\n" );
}


/*---------------------------------------------------------------------------*
  Name        : ReqSendData
  Description : Request send data.
  Arguments   : destCID    - Destination ConnectionID.
              : *buffer    - Send data buffer.
              : length     - Send data size.
              : isReliable - (true)reliable, (false)unreliable.
  Returns     : true  - The send data request was successful.
              : false - The send data request failed.
 *---------------------------------------------------------------------------*/
bool NetworkInfra::ReqSendData( u32 destCID, void *buffer, u32 length, bool isReliable )
{
    bool ret = false;

// Critical section -------->
    m_SendCC.Lock();

    {
        std::memcpy( m_SendDataInfo[m_ReqSendDataIndxW].send_buff, buffer, length );
        m_SendDataInfo[m_ReqSendDataIndxW].destCid   = destCID;
        m_SendDataInfo[m_ReqSendDataIndxW].dataSize  = length;
        m_SendDataInfo[m_ReqSendDataIndxW].reliable  = isReliable;

        // Update...
        m_ReqSendDataIndxW = ( m_ReqSendDataIndxW + 1 ) % NET_SEND_DATA_MAX;
        ret = true;
    }

    m_SendCC.Unlock();
// <-------- Critical section

    return ret;
}

/*---------------------------------------------------------------------------*
  Name        : GetNetInfo
  Description : Get net information.
  Arguments   : *info - Net information buffer.
  Returns     : true  - Session starts.
              : false - Session does not start.
 *---------------------------------------------------------------------------*/
bool NetworkInfra::GetNetInfo( NetInformationStructure *info ) const
{
    if ( m_SessionStartFlg )
    {
        info->hostAddr      = m_Info.hostAddr;
        info->netMask       = m_Info.netMask;
        info->dns1          = m_Info.dns1;
        info->dns2          = m_Info.dns2;
        info->gateway       = m_Info.gateway;
        for ( int i = 0; i < NET_MAX_CONNECT; i++ )
        {
            info->CIDList[i] = m_Info.CIDList[i];
        }
        info->connectionNum = m_Info.connectionNum;
        info->ownerPid      = m_Info.ownerPid;
        info->myPid         = m_Info.myPid;
        info->myCid         = m_Info.myCid;
#ifdef VCT_PLATFORM_CTR
        std::memcpy( info->mac, m_Info.mac, sizeof(u8)*nn::socket::ETH_ALEN );
#else
        //Cafe does not allow acquisition of the MAC address.
#endif
    }

    return m_SessionStartFlg;
}


/*---------------------------------------------------------------------------*
  Name        : CheckNetError
  Description : Check for the network error.
  Arguments   : -
  Returns     : true  - Error.
              : false - No error.
 *---------------------------------------------------------------------------*/
bool NetworkInfra::CheckNetError( void )
{
    return m_NetErrorFlg;
}


/*---------------------------------------------------------------------------*
  Name        : CheckConnection
  Description : Check the network connection.
  Arguments   : -
  Returns     : true  - Connected.
              : false - Disconnected.
 *---------------------------------------------------------------------------*/
bool NetworkInfra::CheckConnection( void )
{
    return m_SessionStartFlg;
}


/*---------------------------------------------------------------------------*
  Name        : NetThread
  Description : Infra thread.
  Arguments   : *net - Pointer of the NetworkInfra class.
  Returns     : -
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
void NetThread( NetworkInfra* net )
{
    net->NetThreadImpl();
}
#else
int NetThread( int,void* net )
{
    SAMPLE_PRINT("NetThread Start.\n");
    ((NetworkInfra *)net)->NetThreadImpl();
    SAMPLE_PRINT("NetThread End.\n");
    return 0x1234;
}
#endif

/*************************** Pia *****************************/
bool PiaNetworkUtil::InitializeCommon(void)
{
    nn::Result result;

    // Variables
    m_ErrorType = NO_ERROR;  //Variable that holds errors that occurred most recently.
    m_NetworkErrorResult = nn::ResultSuccess();
    m_JoinedGatheringId = INVALID_GATHERINGID; // Gathering ID to use for matchmaking.
    m_OwnerPID = INVALID_PRINCIPALID; // Specify the principal ID of the host.
    m_MyPID = INVALID_PRINCIPALID;
    m_IsSessionMaster = 0;  //Is the local system the host?
    m_pNgsFacade = NULL;
    m_pMatchMakingClient = NULL;

    // Render System
    nn::pia::framework::Framework::CreateInstance();
    nn::pia::framework::Framework::GetInstance()->SetFontColor(nn::pia::framework::COLOR_WHITE);
    nn::pia::framework::Framework::GetInstance()->SetFramePerSecond(60);

    // common initialization.
    result = nn::pia::common::Initialize(m_PiaMemoryBlock, MEMORY_BLOCK_SIZE_PIA);
    if(result.IsFailure()){
        SAMPLE_PRINT("pia::common::Initialize Failed.");
        return false;
    }
    result = nn::pia::common::BeginSetup();
    if(result.IsFailure()){
        SAMPLE_PRINT("pia::common::BeginSetup Failed.");
        return false;
    }
    result = nn::pia::common::Scheduler::CreateInstance();
    if(result.IsFailure()){
        SAMPLE_PRINT("pia::common::Scheduler::CreateInstance Failed.");
        return false;
    }
    result = nn::pia::common::EndSetup();
    if(result.IsFailure()){
        SAMPLE_PRINT("pia::common::EndSetup Failed.");
        return false;
    }
    return true;
}
void PiaNetworkUtil::FinalizeCommon(void)
{
    nn::pia::common::Scheduler::DestroyInstance();
    nn::pia::common::Finalize();
    nn::pia::framework::Framework::DestroyInstance();
}

/*!
    @brief This function sets the error reason before jumping to the error state.
 */
void PiaNetworkUtil::SetCommunicationErrorType(SampleErrorType error)
{
    m_ErrorType = error;
}

// Pass the Result of errors that occur when connecting to the Internet.
// The Result passed here is passed to the API function for conversion to a network error code, and is then displayed.
// 
void PiaNetworkUtil::SetNetworkErrorResult(nn::Result result)
{
    m_NetworkErrorResult = result;
}

bool PiaNetworkUtil::InitializeInet(void)
{
    nn::Result result;

    // Internet Communications Preparation Process
    result = Inet_InitializeNetwork();
    if( result.IsFailure() )
    {
#ifdef VCT_PLATFORM_CTR
        if( result == nn::ac::ResultWifiOff() )
        {
            SetCommunicationErrorType( INITIALIZE_ERROR_WIRELESS_OFF );
        }
        else
        {
            SetCommunicationErrorType( INITIALIZE_ERROR_OTHER_REASON );
        }
#else
        SetCommunicationErrorType( INITIALIZE_ERROR_OTHER_REASON );
#endif
        return false;
    }

    // Start processing NEX-related initialization.
    result = Nex_InitializeNex(
        MEMORY_BLOCK_SIZE_NEX,
        MEMORY_BLOCK_SIZE_NEX_EMERGENCY,
        SAMPLE_NEX_THREAD_MODE,
        false);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Allocate instances required for NEX game server-related processing.
    if( m_pNgsFacade == NULL && m_pMatchMakingClient == NULL )
    {
        m_pNgsFacade            = qNew nn::nex::NgsFacade();
        m_pMatchMakingClient    = qNew nn::nex::MatchmakeExtensionClient();
    }
    else
    {
        PIA_PANIC("Not NULL pointer. NgsFacade:%X MatchMakingClient:%X\n", m_pNgsFacade, m_pMatchMakingClient );
    }

    // Started the processing to log in to the server.

    result = Inet_Login(m_pNgsFacade, GAME_ID, ACCESS_KEY, 30 * 1000);
    if (result.IsFailure())
    {
        // The processing to log in failed.
        SetCommunicationErrorType( INITIALIZE_ERROR_OTHER_REASON );
        return false;
    }

    // If logging in succeeds, it does not become NULL.
    PIA_SAMPLE_ASSERT(m_pNgsFacade->GetCredentials() != NULL);

    // Sets the certificate for using the matchmaking service.
    if( !m_pMatchMakingClient->Bind(m_pNgsFacade->GetCredentials()) )
    {
        // Failed to set the certificate.
        SetCommunicationErrorType( INITIALIZE_ERROR_OTHER_REASON );
        return false;
    }

    // Started processing PiaInet initialization.
    result = Inet_InitializePiaInet();
    if (result.IsFailure())
    {
        SetCommunicationErrorType( INITIALIZE_ERROR_OTHER_REASON );
        return false;
    }

    return true;
}

bool PiaNetworkUtil::AutoMatch(u32 maxConnect,u32 netMatchMode)
{
    nn::Result result;
    result = Inet_AutoMatchmake(
        m_pMatchMakingClient, 
        &m_JoinedGatheringId,
        &m_OwnerPID,
        maxConnect,
        netMatchMode,
        "voice chat session"
    );
    if(result.IsFailure())
    {
        return false;
    }
    m_MyPID = m_pMatchMakingClient->GetCredentials()->GetPrincipalID();
    if(m_MyPID == m_OwnerPID)
    {
        m_IsSessionMaster = 1;
    }
    else
    {
        m_IsSessionMaster = 0;
    }
    return true;
}

nn::Result PiaNetworkUtil::NexFacadeStartup(void)
{
    nn::pia::common::CallContext startupContext;
    nn::Result result = nn::pia::inet::NexFacade::GetInstance()->Startup(
        m_pNgsFacade,
        m_pMatchMakingClient,
        GAME_ID );
    if( result.IsSuccess() )
    {
        result = nn::pia::inet::NexFacade::GetInstance()->StartNatSession(&startupContext);
        if( result.IsSuccess() )
        {
            // Waits for NexFacade::StartNatSession to end.
            while( !startupContext.IsCompleted() && result.IsSuccess() )
            {
                nn::nex::Scheduler::GetInstance()->DispatchAll();

                nn::pia::common::Scheduler::GetInstance()->Dispatch();

                result = nn::pia::transport::Transport::GetInstance()->GetConnectionStatus();

#ifdef VCT_PLATFORM_CTR
                nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(15));  // Opens up an appropriate amount of time. (Substitute for VSync.)
#else
                OSSleepMilliseconds(15);  // Opens up an appropriate amount of time. (Substitute for VSync.)
#endif
            }

            if ( result.IsFailure() )
            {
                // Connection is in an abnormal state.
                SAMPLE_PRINT("Transport::GetConnectionStatus became abnormal when waiting StartNatSession process.\n");

                // Implement StopNatSession because there could be an interruption in the middle of StartNatSession asynchronous processing.
                startupContext.Cancel();
                nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                nn::pia::inet::NexFacade::GetInstance()->Cleanup();
                return result;
            }

            result = startupContext.GetResult();

            if( result.IsSuccess() )
            {
                // Transmit the matchmaking session GatheringID to NexFacade.
                nn::pia::inet::NexFacade::GetInstance()->UpdateGatheringID(m_JoinedGatheringId);
            }
            else
            {
                // The asynchronous process of StartNatSession failed.
                SAMPLE_PRINT("NexFacade::StartNatSession process failed.\n");

                nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
                nn::pia::inet::NexFacade::GetInstance()->Cleanup();
            }
        }
        else
        {
            // StartNatSession failed.
            SAMPLE_PRINT("NexFacade::StartNatSession failed.\n");

            nn::pia::inet::NexFacade::GetInstance()->StopNatSession();
            nn::pia::inet::NexFacade::GetInstance()->Cleanup();
        }
    }
    else
    {
        // Startup failed.
        SAMPLE_PRINT("NexFacade::Startup failed.\n");
        nn::pia::inet::NexFacade::GetInstance()->Cleanup();
    }

    return result;
}

bool PiaNetworkUtil::OpenParticipation(void)
{
    if( m_JoinedGatheringId == INVALID_GATHERINGID )
    {
        // The GatheringId that should have been saved is odd.
        return false;
    }

    nn::Result result = Inet_OpenParticipation(m_pMatchMakingClient, m_JoinedGatheringId);
    if (result.IsFailure()) {
        return false;
    }
    return true;
}

bool PiaNetworkUtil::Session(void)
{
    nn::Result result;
    // * The Factory instance is not required except for initialization, so it is allocated with an auto variable.
    nn::pia::inet::NexNetworkFactory  nexFactory;
    // These parameters are borrowed from the transport demo.
    // For more information about the parameters, see the Programming Manual.
    nn::pia::transport::Transport::Setting transportSetting;
    transportSetting.pFactory = &nexFactory;
    transportSetting.maxStationNum              = NET_MAX_CONNECT;
    transportSetting.maxSendThreadBufferNum     = 16;
    transportSetting.maxReceiveThreadBufferNum  = 32;

    result = nn::pia::transport::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::Transport::CreateInstance(transportSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //The port number is hard-coded to 0.
    s_UnreliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::UnreliableProtocol>( 0 );
    result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::UnreliableProtocol>(s_UnreliableProtocolHandle)->Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    s_ReliableProtocolHandle = nn::pia::transport::Transport::GetInstance()->CreateProtocol<nn::pia::transport::ReliableProtocol>( 0 );
    const u32 SEND_BUF_NUM = 64;
    const u32 RECV_BUF_NUM = 64;
    result = nn::pia::transport::Transport::GetInstance()->GetProtocol<nn::pia::transport::ReliableProtocol>(s_ReliableProtocolHandle)->Initialize(SEND_BUF_NUM,RECV_BUF_NUM);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::transport::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //Priority settings for the send/receive thread. (For information about priorities, see SampleConfig.h.)
    //The priority of the send/receive thread does not necessarily need to be changed, but in this sample, the priority has been changed to show the change method.
    //Priority has been changed.
    nn::pia::transport::ThreadStreamManager* pThreadStreamManager = nn::pia::transport::ThreadStreamManager::GetInstance();
    PIA_SAMPLE_ASSERT (pThreadStreamManager); //If the transport initialization has finished, the instance has probably been created.
#ifdef VCT_PLATFORM_CTR
    pThreadStreamManager->SetSendThreadPriority(nn::os::DEFAULT_THREAD_PRIORITY -1);
    pThreadStreamManager->SetReceiveThreadPriority(nn::os::DEFAULT_THREAD_PRIORITY -2);
#else
    pThreadStreamManager->SetSendThreadPriority(OS_PRIORITY_APP_DEFAULT -1);
    pThreadStreamManager->SetReceiveThreadPriority(OS_PRIORITY_APP_DEFAULT -2);
#endif

    //Initialize nn::pia::session.
    nn::pia::transport::StationConnectionInfo stationConnectionInfo;
    nn::pia::common::CallContext callContext;

    // These parameters are borrowed from the transport demo.
    // For more information about the parameters, see the Programming Manual.
    nn::pia::session::Session::Setting sessionSetting;
    sessionSetting.pNetworkFactory = &nexFactory;
    sessionSetting.networkTopology   = nn::pia::session::NetworkTopology_FullMesh;
    sessionSetting.hostMigrationMode = nn::pia::session::HostMigrationMode_None;

    result = nn::pia::session::Initialize();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::BeginSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::Session::CreateInstance(sessionSetting);
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    result = nn::pia::session::EndSetup();
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    // Configure the packet with the signature feature (Internet).
    // Convert the shared session key obtained from the server during matchmaking so that it can be used as the signature key.
    nn::pia::common::SignatureSetting *pSignatureSetting = Inet_GetMatchmakeSessionSignatureSetting();
    PIA_SAMPLE_ASSERT(pSignatureSetting);
    result = Session_StartupSession( *(pSignatureSetting),NULL );
    PIA_SAMPLE_ASSERT_RESULT_IS_SUCCESS(result);

    //Registers the event callback.
    //Before callback registration, initialize the structure that the callback accesses.
    for(int i=0; i< NET_MAX_CONNECT ; i++)
    {
        s_NodeStatus[i].stationId               = nn::pia::STATION_ID_INVALID; //Only shows whether it is valid.
    }
    nn::pia::session::Session::GetInstance()->RegisterSessionEventCallback( sessionCallback );

    // NexFacade setup.
    result = NexFacadeStartup();
    if ( result.IsFailure() )
    {
        if ( result == nn::pia::inet::ResultDnsFailed() )
        {
            // DNS resolution failed.
            SAMPLE_PRINT("SceneInetMatchmake:NexFacadeStartup failed. (DnsFailed)\n");
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
        }
        else if ( result == nn::pia::inet::ResultNatCheckFailed() )
        {
            // NAT check failed
            SAMPLE_PRINT("SceneInetMatchmake:NexFacadeStartup failed. (NatCheckFailed)\n");
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
        }
        else if ( result == nn::pia::ResultConnectionIsLost() )
        {
            // Connection is in an abnormal state.
            SAMPLE_PRINT("SceneInetMatchmake:NexFacadeStartup failed. (ConnectionIsLost)\n");
            SetCommunicationErrorType( CONNECTION_LOST_ERROR );
        }
        else
        {
            //Unexpected failure.
            SAMPLE_PRINT("SceneInetMatchmake:NexFacadeStartup failed. (Unknown)\n");
            PIA_TRACE_RESULT(result);
            PIA_SAMPLE_ASSERT(false);
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
        }
        SetNetworkErrorResult( result );
        return false;
    }

    if( m_IsSessionMaster )
    {
        result = nn::pia::session::Session::GetInstance()->CreateSessionAsync();
        if( result.IsFailure() )
        {
            SAMPLE_PRINT("SceneInetMatchmake:CreateSessionAsync failed.\n");
            SetNetworkErrorResult( result );
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
            return false;
        }
    }
    else
    {
        //Gets the StationConnectionInfo for the session host.
        //Because this process takes time, display a message such as "Please wait..."

        // Gets the StationURL of the session host.
        nn::nex::qList<nn::nex::StationURL> lstURLs;

        result = Inet_GetHostStationURLs(m_pMatchMakingClient, m_JoinedGatheringId, &lstURLs);
        if (result.IsFailure())
        {
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
            SetNetworkErrorResult( SampleFailedGetHostStationConnectionInfo() );
            return false;
        }
        // Succeeded at getting StationURL and converting it to StationConnectionInfo.
        if( nn::pia::inet::NexFacade::ConvertNexStationURLToStationConnectionInfo( lstURLs, &stationConnectionInfo ).IsFailure() )
        {
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
            SetNetworkErrorResult( SampleFailedGetHostStationConnectionInfo() );
            return false;
        }
        result = nn::pia::session::Session::GetInstance()->JoinSessionAsync( stationConnectionInfo );
        if( result.IsFailure() )
        {
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
            SetNetworkErrorResult( result );
            return false;
        }
    }

    // Cycle a render loop because JoinSession takes time and nn::pia::common::Scheduler::Dispatch is required.
    // The following executes only when joining a session.
    while(1)
    {
        // The condition determinations for exiting a loop differ between the host and client.
        if(m_IsSessionMaster)
        {
            if(nn::pia::session::Session::GetInstance()->IsCompletedCreateSession())
            {
                break;
            }
        }
        else
        {
            if(nn::pia::session::Session::GetInstance()->IsCompletedJoinSession())
            {
                break;
            }
        }

        // To maintain communication with the NEX game server, you must call nn::nex::Scheduler::Dispatch.
        // In this demo, it is called in every frame, but you may reduce the call frequency to reduce processing.
        nn::nex::Scheduler::GetInstance()->DispatchAll();
        nn::pia::common::Scheduler::GetInstance()->Dispatch();

        //Status check for the network connection.
        result = nn::pia::transport::Transport::GetInstance()->GetConnectionStatus();
        if ( result.IsFailure() )
        {
            //If the network connection is abnormal:
            if ( result == nn::pia::ResultConnectionIsLost() )
            {
                //Returns when the network connection is interrupted because the wireless switch is turned off or communication fails.
                SAMPLE_PRINT("Connection Lost when waiting Join process.\n");
            }
            else
            {
                //Unexpected state.
                SAMPLE_PRINT("Transport::GetConnectionStatus is unknown status when waiting Join process.\n");
                PIA_TRACE_RESULT(result);
                PIA_SAMPLE_ASSERT(false);
            }

            // Call Cleanup because an error occurred during Session processing.
            nn::pia::session::Session::GetInstance()->Cleanup();

            SetCommunicationErrorType( CONNECTION_LOST_ERROR );
            SetNetworkErrorResult( result );
            return false;
        }

        // The main thread is going from Begin through EndFrame, so this cannot be used here.
#ifdef VCT_PLATFORM_CTR
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(16));
#else
        OSSleepMilliseconds(16);
#endif
    }

    // Gets the results of an asynchronous process (the API function to call differs between host and client).
    if( m_IsSessionMaster )
    {
        result = nn::pia::session::Session::GetInstance()->GetCreateSessionResult();
    }
    else
    {
        result = nn::pia::session::Session::GetInstance()->GetJoinSessionResult();
    }

    //Confirm the result for asynchronous processing.
    if( result.IsFailure() )
    {
        if ( result == nn::pia::ResultInvalidState() )
        {
            // It was not a state where connection was possible.
            SAMPLE_PRINT("Join process Failed (InvalidState)\n");
        }
        else if ( result == nn::pia::ResultTimeOut() )
        {
            // The process timed out.
            SAMPLE_PRINT("Join process Failed (TimeOut)\n");
        }
        else if ( result == nn::pia::session::ResultStationConnectionFailed() )
        {
            // Failed to connect.
            SAMPLE_PRINT("Join process Failed (StationConnectionFailed)\n");
        }
        else if ( result == nn::pia::session::ResultJoinRequestDenied() )
        {
            // The session host refused the connection.
            SAMPLE_PRINT("Join process Failed (JoinRequestDenied)\n");
        }
        else if ( result == nn::pia::session::ResultSessionIsFull() )
        {
            // The session that you wanted to participate in was full, so you could not participate.
            SAMPLE_PRINT("Join process Failed (SessionIsFull)\n");
        }
        else if ( result == nn::pia::session::ResultInvalidSystemMessage() )
        {
            // Because an invalid system message was received, the process was ended.
            SAMPLE_PRINT("Join process Failed (InvalidSystemMessage)\n");
        }
        else if ( result == nn::pia::transport::ResultIncompatibleFormat() )
        {
            // The communication format of the peer is incompatible. Could not connect.
            SAMPLE_PRINT("Join process Failed (IncompatibleFormat)\n");
        }
        else if ( result == nn::pia::session::ResultFailedStartupProtocol() )
        {
            // The system protocol start process failed.
            SAMPLE_PRINT("Create process Failed (FailedStartupProtocol)\n");
        }
        else
        {
            // Any other errors.
            SAMPLE_PRINT("Join process Failed (Unknown)\n");
            PIA_TRACE_RESULT(result);
        }

        // Treated as a matchmaking error.
        SetCommunicationErrorType( MATCH_MAKE_ERROR );
        SetNetworkErrorResult( result );
        return false;
    }

    if ( m_IsSessionMaster )
    {
        // Use OpenParticipation because the Session (P2P communication) preparations are finished.
        //Because this process takes time, display a message such as "Please wait..."
        if( !OpenParticipation() )
        {
            PIA_CACHED_PRINTF("ScenePrepareSession:OpenParticipation failed.\n");
            SetNetworkErrorResult( SampleFailedOpenParticipation() );
            SetCommunicationErrorType( MATCH_MAKE_ERROR );
            return false;
        }
    }

    return true;
}

void PiaNetworkUtil::FinalizeInet(void)
{
    // PiaInet finalization processing.
    Inet_FinalizePiaInet();

    if( m_pMatchMakingClient != NULL )
    {
        if( m_JoinedGatheringId != INVALID_GATHERINGID )
        {
            // Finishes the matchmaking session.
            Inet_LeaveMatchmakeSession(
                m_pMatchMakingClient,
                m_JoinedGatheringId,
                m_OwnerPID,
                true
            );

            m_JoinedGatheringId = INVALID_GATHERINGID;
        }

        // Finalizes the matchmaking service.
        m_pMatchMakingClient->Unbind();
        qDelete m_pMatchMakingClient;
        m_pMatchMakingClient = NULL;
    }

    // Process to log out from the server.
    if (m_pNgsFacade != NULL) {
        Inet_Logout(m_pNgsFacade);
        qDelete m_pNgsFacade;
        m_pNgsFacade = NULL;
    }

    // Processes NEX-related finalization.
    Nex_FinalizeNex();

    // Processes Internet finalization (disconnection).
    Inet_FinalizeNetwork();
}
