/*---------------------------------------------------------------------------*
  Project:  Voice Chat Sample.
  File:     sndmic.cpp

  Copyright (C) 1998-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*
 * Includes
 *---------------------------------------------------------------------------*/
#include "framework.h"
#include "sndmic.h"
#include "utilities.h"


/*---------------------------------------------------------------------------*
    Const
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#else
static const u16 speakerGainTbl[SPEAKER_GAIN_IDX_MAX] =
{
        0, // 0.0
     3277, // 0.1
     6554, // 0.2
     9831, // 0.3
    13108, // 0.4
    16384, // 0.5
    19661, // 0.6
    22938, // 0.7
    26215, // 0.8
    29492, // 0.9
    32768, // 1.0
};
#endif

/*---------------------------------------------------------------------------*
 *  Define
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#define ADD_BUFFER_NUM          (2)         // Num of add  buffer
#define SOUND_FREQ_8KHZ         (NN_SND_HW_I2S_CLOCK_32KHZ/4)   //  8kHz sampling(8182)
//#define FREQ8KHZ_TIME_NS        (1000000000/SOUND_FREQ_8KHZ)    // 122217.646354288786 nsec
#define FREQ8KHZ_TIME_NS        (1000000000/SOUND_FREQ_8KHZ - 0.89) // Fine adjustment. Not synchronized with the microphone sampling.
#define MIC_CAPTURE_MARGIN_MS   (12)
#define MIC_CAPTURE_MARGIN_SAMPLE (MIC_CAPTURE_MARGIN_MS * VCT_AUDIO_DATA_SIZE_PER_MILLI_SEC / 2)
#define ROMFILE_WAVE_0          (L"rom:/noise.wav")
#define ROMFILE_WAVE_1          (L"rom:/busy.wav")
#define ROMFILE_WAVE_2          (L"rom:/kart_title.wav")
#else
// Num of stream buffer blocks.
#define STREAMBUFFER_BLOCKS     ( 16 )
// Num of voice buffer blocks.
// AX unit is 3 ms -> Set to a multiple of 3 to skip copy processing that crosses ring buffer boundaries.
#define VOICEBUFFER_BLOCKS      ( 3 )
// AX voice stream buffer size.
#define STREAMBUFFER_SIZE       ( AX_IN_SAMPLES_PER_FRAME * sizeof( s16 ) * STREAMBUFFER_BLOCKS )
// VoiceChat data buffer size.
#define VOICEBUFFER_BASE_SIZE   ( VOICEBUFFER_BLOCKS * sizeof( s16 ) )
// Audio data file path.
#define BUSY_FILE_PATH          ( "vct/busy.pcm" )
#define NOISE_FILE_PATH         ( "vct/noise.pcm" )
#define MIC_RING_BUFFER_SIZE_BYTES (20*1024)
#define MIC_LEVEL_SCALE         (100)    // Max value of mic gain level
#endif

/*---------------------------------------------------------------------------*
 * Global Variables
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
#else
SndMicDevice* SndMicDevice::m_This;
#endif

/*---------------------------------------------------------------------------*
    Prototype
 *---------------------------------------------------------------------------*/
#ifdef VCT_PLATFORM_CTR
void SoundThreadFunc( SndMicDevice* sndmicDev );
void FrameProcessThreadFunc( SndMicDevice* sndmicDev );
#endif

#ifdef VCT_PLATFORM_CTR

/*---------------------------------------------------------------------------*
  Name: Initialize
  Description: Initialize Snd & Mic Device
  Arguments: *config      - sndmic config parameter.
  Returns: SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SNDMIC_ERROR_PARAM - parameter error.
              :   SND_ERROR_DSP      - dsp error.
              :   MIC_ERROR_LIB      - mic library error.
              :   SND_ERROR_LIB      - snd library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Initialize( SndMicConfigStructure *config )
{
    nn::pia::framework::Framework* pFw = nn::pia::framework::Framework::GetInstance();
    nn::Result          Result;
    nn::snd::MixParam   Mix;
    nn::fs::FileReader  fileReader;
    WaveFmt             fmt;

    /* Check argument */
    if (( pFw == NULL ) || ( config == NULL ))
    {
        return SNDMIC_ERROR_PARAM;
    }

    /* Config parameter copy */
    m_Config.frame_msec     = config->frame_msec;
    m_Config.mic_gain      = config->mic_gain;

    /* frame size... */
    m_AudioFrameSize   = m_Config.frame_msec * 8 * sizeof(s16);

    if ( m_AudioFrameSize < 1056 )
    {
        m_SoundBufferSize  = 1056;
    }
    else
    {
        m_SoundBufferSize  = m_AudioFrameSize;
    }

    /* Initialize variables... */
    m_SoundStatus              = SND_PROCESS_NONE;
    m_CurrBufferIdx            = 0;
    m_SoundBufferReqIdx        = 0;
    m_MicTimingError           = false;
    m_Pause                    = 0;

    m_MicStatus                = MIC_PROCESS_NONE;

    m_MicSyncEvent.Initialize(false);
    m_MicSyncEvent.ClearSignal();
    m_MicStopEvent.Initialize(false);
    m_MicStopEvent.ClearSignal();

    /**********   MIC Initialization   **********/
    if ( nn::mic::Initialize().IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* MIC-Buffer == 4096*2byte */
    if ( nn::mic::SetBuffer(m_MicBuffer, sizeof(m_MicBuffer)).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Set MIC-Amp. */
    if ( nn::mic::GetAmpGain(&m_MicGain).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    if ( nn::mic::SetAmpGain(m_Config.mic_gain).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Power On MIC. */
    if ( nn::mic::SetAmp( true ).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /***** Update process status *****/
    m_MicStatus = MIC_PROCESS_INIT;


    /**********   SND Initialization   **********/
    /* Initialize DSP + Load DSP Component */
    Result = nn::dsp::Initialize();
    if ( Result.IsFailure() )
    {
        return SND_ERROR_DSP;
    }

    Result = nn::dsp::LoadDefaultComponent();
    if ( Result.IsFailure() )
    {
        return SND_ERROR_DSP;
    }

    Result = nn::snd::Initialize();
    if ( Result.IsFailure() )
    {
        return SND_ERROR_LIB;
    }


    /* Prepare Sound Buffer */
    for ( int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++ )
    {
        /* Allocate SoundBuffer (as described in Programming Manual, SoundBuffer must be at least 1025 bytes) */
        m_SoundBuffer[Idx] = reinterpret_cast<u8*>( pFw->AllocateDeviceMemory( m_SoundBufferSize ) );
        std::memset( m_SoundBuffer[Idx], 0x0, m_SoundBufferSize );
        nn::snd::FlushDataCache( reinterpret_cast<uptr>( m_SoundBuffer[Idx] ), m_SoundBufferSize );
    }

    /* Prepare WaveBuffer */
    for ( int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++ )
    {
        /* Register to WaveBuffer */
        nn::snd::InitializeWaveBuffer( &m_WaveBuffer[Idx] );
        m_WaveBuffer[Idx].bufferAddress = reinterpret_cast<void*>( m_SoundBuffer[Idx] );
        m_WaveBuffer[Idx].sampleLength  = nn::snd::GetSampleLength( m_AudioFrameSize, nn::snd::SAMPLE_FORMAT_PCM16, 1 );
        m_WaveBuffer[Idx].loopFlag      = false;
    }

    /* Prepare volume Mix */
    Mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT]  = 1.0f;
    Mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 1.0f;

    /* Prepare Voice */
    m_pVoice = nn::snd::AllocVoice( 128, NULL, NULL );
    m_pVoice->SetSampleRate( SOUND_FREQ_8KHZ );                /* Set sampling rate (8 kHz)  */
    m_pVoice->SetSampleFormat( nn::snd::SAMPLE_FORMAT_PCM16 ); /* Specify format (PCM16)       */
    m_pVoice->SetChannelCount( 1 );                            /* Set number of channels (mono)          */
    m_pVoice->SetPitch( 1.0f );                                /* Set pitch                    */
//    m_pVoice->SetVolume( 10.0f );                              /* Set volume                */
    m_pVoice->SetVolume( 1.0f );                               /* Volume settings.                */
    m_pVoice->SetMixParam( Mix );                              /* Set volume Mix           */
    for ( int Idx = 0; Idx < ADD_BUFFER_NUM; Idx++ )
    {
        m_pVoice->AppendWaveBuffer( &m_WaveBuffer[Idx] );      /* Add buffer.                  */
    }

/* Prepare background music */
    /* Load background music data */
    Result = fileReader.TryInitialize( ROMFILE_WAVE_2 );
    NN_UTIL_PANIC_IF_FAILED( Result );
    m_pWaveMemory[2] = reinterpret_cast<u8*>( pFw->AllocateDeviceMemory( GetWaveLength( &fileReader ) + m_AudioFrameSize ) );
    if ( m_pWaveMemory[2] == NULL )
    {
        SAMPLE_PRINT( "Error:Allocate memory error.\n" );
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[2].buf = m_pWaveMemory[2];
    LoadWave( &fileReader, &fmt, &m_WaveData[2] );
    fileReader.Finalize();

    nn::snd::FlushDataCache( reinterpret_cast<uptr>( m_WaveData[2].buf ), m_WaveData[2].size );

    /* Register to WaveBuffer */
    nn::snd::InitializeWaveBuffer( &m_MusicWaveBuffer );
    m_MusicWaveBuffer.bufferAddress = reinterpret_cast<void*>( m_WaveData[2].buf );
    m_MusicWaveBuffer.sampleLength  = nn::snd::GetSampleLength( m_WaveData[2].size, nn::snd::SAMPLE_FORMAT_PCM16, 1 );
    m_MusicWaveBuffer.loopFlag      = true;

    /* Prepare voice for background music */
    m_pMusic = nn::snd::AllocVoice( 128, NULL, NULL );
    m_pMusic->SetSampleRate( SOUND_FREQ_8KHZ );                /* Set sampling rate (8 kHz)  */
    m_pMusic->SetSampleFormat( nn::snd::SAMPLE_FORMAT_PCM16 ); /* Specify format (PCM16)       */
    m_pMusic->SetChannelCount( 1 );                            /* Set number of channels (mono)        */
    m_pMusic->SetPitch( 1.0f );                                /* Set pitch                    */
    m_pMusic->SetVolume( 0.2f );                               /* Volume settings.                */
    m_pMusic->SetMixParam( Mix );                              /* Set volume Mix           */
    m_pMusic->AppendWaveBuffer( &m_MusicWaveBuffer );          /* Add buffer.                  */

    /* Set master volume. */
    nn::snd::SetMasterVolume( 1.0f );

    /* Load noise data */
    Result = fileReader.TryInitialize( ROMFILE_WAVE_0 );
    NN_UTIL_PANIC_IF_FAILED( Result );
    m_pWaveMemory[0] = reinterpret_cast<u8*>( pFw->AllocateDeviceMemory( GetWaveLength( &fileReader ) + m_AudioFrameSize) );
    if ( m_pWaveMemory[0] == NULL )
    {
        SAMPLE_PRINT( "Error:Allocate memory error.\n" );
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[0].buf = m_pWaveMemory[0];
    LoadWave( &fileReader, &fmt, &m_WaveData[0] );
    fileReader.Finalize();
    m_FrameCount[0] = static_cast<u32>( (m_WaveData[0].size + m_AudioFrameSize - 1) / m_AudioFrameSize );

    /* Load busy data */
    Result = fileReader.TryInitialize( ROMFILE_WAVE_1 );
    NN_UTIL_PANIC_IF_FAILED( Result );
    m_pWaveMemory[1] = reinterpret_cast<u8*>( pFw->AllocateDeviceMemory( GetWaveLength( &fileReader ) + m_AudioFrameSize) );
    if ( m_pWaveMemory[1] == NULL )
    {
        SAMPLE_PRINT( "Error:Allocate memory error.\n" );
        return SNDMIC_ERROR_NONE;
    }
    m_WaveData[1].buf = m_pWaveMemory[1];
    LoadWave( &fileReader, &fmt, &m_WaveData[1] );
    fileReader.Finalize();
    m_FrameCount[1] = static_cast<u32>( (m_WaveData[1].size + m_AudioFrameSize - 1) / m_AudioFrameSize );

    m_FramePlayReq = false;

    /***** Update process status *****/
    m_SoundStatus = SND_PROCESS_INIT;


    /****************************** Thread Registration *************************/
    /* Register DSP callback function */
    m_SoundThread.StartUsingAutoStack( SoundThreadFunc, this, (int)4096, (int)2 );

    /* Register thread for processing frame data */
    m_FrameProcessThread.StartUsingAutoStack( FrameProcessThreadFunc, this, (int)4096, (int)(NN_OS_DEFAULT_THREAD_PRIORITY - 2));



    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : Start
  Description : Start playing stream.
  Arguments   : -
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Start(void)
{
    nn::Result Result;

    /* Start mic using FrameProcess thread. */
    /* Update status when changing from STOP to START to match start timing. */
    if( m_MicStatus == MIC_PROCESS_STOP )
    {
        m_MicStatus = MIC_PROCESS_INIT;
    }


    /**********   SND Start **********/
    /* Check process status */
    if ( ( m_SoundStatus != SND_PROCESS_INIT ) && ( m_SoundStatus != SND_PROCESS_STOP ) )
    {
        return SND_ERROR_SEQ;
    }

    /* Start playback */
    m_pVoice->SetState( nn::snd::Voice::STATE_PLAY );

    /* Update process status */
    m_SoundStatus = SND_PROCESS_PLAY;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : Stop
  Description : Stop playing stream.
  Arguments   : -
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Stop(void)
{

    /**********   MIC Stop   **********/
    /* Check process status */
    if ( m_MicStatus != MIC_PROCESS_START )
    {
        return MIC_ERROR_SEQ;
    }

    /* Update process status */
    m_MicStatus = MIC_PROCESS_STOP;
    /* Wait until processing ends. */
    m_MicStopEvent.Wait();


    /**********   SND Stop   **********/
    /* Check process status */
    if ( m_SoundStatus != SND_PROCESS_PLAY )
    {
        return SND_ERROR_SEQ;
    }

    /* End playback */
    if ( m_pVoice->GetState() == nn::snd::Voice::STATE_PLAY )
    {
        m_pVoice->SetState( nn::snd::Voice::STATE_STOP );
    }

    /* Update process status */
    m_SoundStatus = SND_PROCESS_STOP;


    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : SetVolume
  Description : Set voice volume.
  Arguments   : volume - value of volume.
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetVolume( f32 volume )
{
    nn::Result Result;

    /* Check process status */
    if ( m_SoundStatus != SND_PROCESS_PLAY )
    {
        return SND_ERROR_SEQ;
    }

    /* Set Volume... */
    if ( m_pVoice != NULL )
    {
        m_pVoice->SetVolume( volume );
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : SetBGMVolume
  Description : Set BGM volume.
  Arguments   : volume - value of BGM volume.
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetBGMVolume( f32 volume )
{
    nn::Result Result;

    /* Check process status */
    if ( m_SoundStatus != SND_PROCESS_PLAY )
    {
        return SND_ERROR_SEQ;
    }

    /* Set Volume... */
    if ( m_pMusic != NULL )
    {
        m_pMusic->SetVolume( volume );
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : StartBGM
  Description : Start playing BGM.
  Arguments   : -
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StartBGM( void )
{
    nn::Result Result;

    /* Check process status */
    if ( m_SoundStatus != SND_PROCESS_PLAY )
    {
        return SND_ERROR_SEQ;
    }

    /* Start background music playback */
    m_pMusic->SetState( nn::snd::Voice::STATE_PLAY );

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : StopBGM
  Description : Stop playing BGM.
  Arguments   : -
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ   - snd sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StopBGM( void )
{
    nn::Result Result;

    /* Check process status */
    if ( m_SoundStatus != SND_PROCESS_PLAY )
    {
        return SND_ERROR_SEQ;
    }

    /* Stop background music playback */
    if ( m_pMusic->GetState() == nn::snd::Voice::STATE_PLAY )
    {
        m_pMusic->SetState( nn::snd::Voice::STATE_PAUSE );
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : Finalize
  Description : Finalize Snd & Mic Device
  Arguments   : -
  Returns     : SNDMIC_ERROR_STATUS
              :   SNDMIC_ERROR_NONE  - none error.
              :   SND_ERROR_SEQ      - snd sequence error.
              :   SNDMIC_ERROR_PARAM - parameter error.
              :   SND_ERROR_DSP      - dsp error.
              :   SND_ERROR_LIB      - snd library error.
              :   MIC_ERROR_SEQ      - mic sequence error.
              :   MIC_ERROR_LIB      - mic library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::Finalize( void )
{
    nn::pia::framework::Framework* pFw = nn::pia::framework::Framework::GetInstance();
    nn::Result  Result;

    /**********   MIC Finalizing   **********/
    /* Check process status */
    if ( ( m_MicStatus != MIC_PROCESS_STOP ) && ( m_MicStatus != MIC_PROCESS_INIT ) )
    {
        return MIC_ERROR_SEQ;
    }

    /* Power off mic. */
    if ( nn::mic::SetAmp(false).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Return mic-amp setting. */
    if ( nn::mic::SetAmpGain(m_MicGain).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Free MIC-Buffer */
    if ( nn::mic::ResetBuffer().IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Finalize MIC */
    if ( nn::mic::Finalize().IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    /* Update process status */
    m_MicStatus = MIC_PROCESS_NONE;

    /**********   Finalize Frame Data Thread   **********/
    m_FrameProcessThread.Join();
    m_FrameProcessThread.Finalize();


    /**********   SND Finalizing   **********/
    /* Check process status */
    if ( ( m_SoundStatus != SND_PROCESS_STOP ) && ( m_SoundStatus != SND_PROCESS_INIT ) )
    {
        return SND_ERROR_SEQ;
    }

    /* Update process status */
    m_SoundStatus = SND_PROCESS_NONE;

    /* Destroy the sound thread */
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    /* Destroy Voice */
    nn::snd::FreeVoice( m_pVoice );
    nn::snd::FreeVoice( m_pMusic );

    /* Destroy sound buffer */
    for ( int Idx = 0; Idx < SOUND_BUFFER_NUM; Idx++ )
    {
        pFw->FreeDeviceMemory( m_SoundBuffer[Idx] );
    }

    /* Finalize SND */
    Result = nn::snd::Finalize();
    if ( Result.IsFailure() )
    {
        return SND_ERROR_LIB;
    }

    /* Finalize DSP */
    Result = nn::dsp::UnloadComponent();
    if ( Result.IsFailure() )
    {
        return SND_ERROR_DSP;
    }
    nn::dsp::Finalize();


    if ( m_pWaveMemory[0] != NULL )
    {
        pFw->FreeDeviceMemory( m_pWaveMemory[0] );
        m_pWaveMemory[0] = NULL;
    }

    if ( m_pWaveMemory[1] != NULL )
    {
        pFw->FreeDeviceMemory( m_pWaveMemory[1] );
        m_pWaveMemory[1] = NULL;
    }

    if ( m_pWaveMemory[2] != NULL )
    {
        pFw->FreeDeviceMemory( m_pWaveMemory[2] );
        m_pWaveMemory[2] = NULL;
    }



    m_MicStopEvent.Finalize();
    m_MicSyncEvent.Finalize();

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : SoundThreadFuncImpl
  Description : Sound Thread function for send Parameter to DSP.(Core)
  Arguments   : -
  Returns     : -
 *---------------------------------------------------------------------------*/
void SndMicDevice::SoundThreadFuncImpl( void )
{
    bool store_req = false;

    while ( true )
    {
        /* Wait for data from DSP*/
        nn::snd::WaitForDspSync();

        if ( m_SoundStatus == SND_PROCESS_NONE )
        {
            break;
        }

        if( m_SoundStatus == SND_PROCESS_PLAY )
        {
            /* Triggering MIC here to make sure that SND and MIC are started at a specific timing as much as possible */
            if ( m_MicStatus == MIC_PROCESS_INIT )
            {
                m_MicSyncEvent.Signal();
            }

            /* Check WaveBuffer status (done playing back -> add buffer) */
            if ( ( m_pVoice != NULL ) &&
                 ( m_WaveBuffer[m_CurrBufferIdx].status == nn::snd::WaveBuffer::STATUS_DONE ) &&
                 ( m_WaveBuffer[( m_CurrBufferIdx + ADD_BUFFER_NUM ) % SOUND_BUFFER_NUM].status == nn::snd::WaveBuffer::STATUS_FREE ))
            {
                /* Re-add buffer */
                m_pVoice->AppendWaveBuffer( &m_WaveBuffer[( m_CurrBufferIdx + ADD_BUFFER_NUM ) % SOUND_BUFFER_NUM] );
                store_req = true;
                m_SoundBufferReqIdx = m_CurrBufferIdx;
                /* Update Curr Buffer Indx */
                m_CurrBufferIdx = ( m_CurrBufferIdx + 1 ) % SOUND_BUFFER_NUM;
            }

            if(store_req)
            {
                /* Get playback data */
                if(GetReceiveAudioData(m_SoundBuffer[m_SoundBufferReqIdx], m_AudioFrameSize))
                {
                    if ( m_FramePlayReq )
                    {
                        if ( m_FramePlayNum < m_FrameCount[m_WaveNumbber] )
                        {
                            std::memcpy( m_SoundBuffer[m_SoundBufferReqIdx], m_pPlayWave, m_AudioFrameSize );
                            m_FramePlayNum++;
                            m_pPlayWave += m_AudioFrameSize;
                        }
                        else
                        {
                            m_FramePlayReq = false;
                        }
                    }

                    /* Flush Cache */
                    nn::snd::FlushDataCache( reinterpret_cast<uptr>( m_SoundBuffer[m_SoundBufferReqIdx] ), m_AudioFrameSize );

                    /* Re-add buffer */
                    nn::snd::InitializeWaveBuffer( &m_WaveBuffer[m_SoundBufferReqIdx] );
                    m_WaveBuffer[m_SoundBufferReqIdx].bufferAddress =
                                                      reinterpret_cast<void*>( m_SoundBuffer[m_SoundBufferReqIdx] );
                    m_WaveBuffer[m_SoundBufferReqIdx].sampleLength  =
                                                      nn::snd::GetSampleLength( m_AudioFrameSize, nn::snd::SAMPLE_FORMAT_PCM16, 1 );
                    m_WaveBuffer[m_SoundBufferReqIdx].loopFlag      = false;

                    store_req = false;
                }
            }

                /* Send reference mixed data... */
            if ( nn::snd::GetMixedBusData( m_MixedData, NN_SND_SAMPLES_PER_FRAME ) )
            {
                VCT_HandleReference( (u8 *)m_MixedData, NN_SND_SAMPLES_PER_FRAME*2*sizeof(s16) );
            }
        }
        /* Send arguments to DSP */
        nn::snd::SendParameterToDsp();
    }
}


/*---------------------------------------------------------------------------*
  Name        : RomfileWavePlay
  Description : Play Romfile wave data.
  Arguments   : number - wave data number.
  Returns     : -
 *---------------------------------------------------------------------------*/
void SndMicDevice::RomfileWavePlay( u32 number )
{
    m_FramePlayReq = true;
    m_FramePlayNum = 0;
    m_WaveNumbber  = number;
    m_pPlayWave    = m_WaveData[number].buf;
}


/*---------------------------------------------------------------------------*
  Name        : SoundThreadFunc
  Description : Sound Thread function for send Parameter to DSP (Wrapper).
  Arguments   : sndmicDev - SndMicDevice Class pointer.
  Returns     : -
 *---------------------------------------------------------------------------*/
void SoundThreadFunc( SndMicDevice* sndmicDev )
{
    sndmicDev->SoundThreadFuncImpl();
}


/*---------------------------------------------------------------------------*
  Name        : MicGainSet
  Description : Change MIC Gain.
  Arguments   : gain - set mic gain value.
  Returns     : SNDMIC_ERROR_STATUS
              :    SNDMIC_ERROR_NONE  - none error.
              :    MIC_ERROR_SEQ   - sequence error.
              :    MIC_ERROR_LIB   - library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::MicGainSet( u8 gain )
{
    /* Check process status */
    if ( m_MicStatus == MIC_PROCESS_NONE )
    {
        return MIC_ERROR_SEQ;
    }

    /* Set mic gain. */
    if ( nn::mic::SetAmpGain( gain ).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : MicGainGet
  Description : Get MIC Gain.
  Arguments   : *gain - set mic gain value.
  Returns     : SNDMIC_ERROR_STATUS
              :    SNDMIC_ERROR_NONE  - none error.
              :    MIC_ERROR_SEQ      - sequence error.
              :    MIC_ERROR_LIB      - library error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::MicGainGet( u8 *gain ) const
{
    /* Check process status */
    if ( m_MicStatus == MIC_PROCESS_NONE )
    {
        return MIC_ERROR_SEQ;
    }

    if ( nn::mic::GetAmpGain( gain ).IsFailure() )
    {
        return MIC_ERROR_LIB;
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
  Name        : GetReceiveAudioData
  Description : Get ReceiveAudio Data.
  Arguments   : *gain - set mic gain value.
  Returns     : true    success
              : false   no data
 *---------------------------------------------------------------------------*/
bool SndMicDevice::GetReceiveAudioData( u8 *audioBuf, u16 length )
{
    /* Check Index */
    if ( m_SndBufReadIdx == m_SndBufWriteIdx )
    {
        return false;
    }
    else
    {
        for(int i = 0;i < length;i++)
        {
            audioBuf[i] = m_SndBuf[m_SndBufReadIdx][i];
        }
        m_SndBufReadIdx = (m_SndBufReadIdx + 1) % 4;

        return true;
    }
}


/*---------------------------------------------------------------------------*
  Name        : GetSndMicErrorFlag
  Description : Get SndMic Error Flag.
  Arguments   : 
  Returns     : true    error occurred
              : false   no error
 *---------------------------------------------------------------------------*/
bool SndMicDevice::GetSndMicErrorFlag( void)
{
    bool ErrorFlag;

    if(m_MicTimingError)
    {
        ErrorFlag = true;
        m_MicTimingError = false;
    }
    else
    {
        ErrorFlag = false;
    }

    return ErrorFlag;
}


/*---------------------------------------------------------------------------*
  Name        : Pause
  Description : pause or restart
  Arguments   : -
  Returns     : 0 restart
              : 1 paused
 *---------------------------------------------------------------------------*/
u8 SndMicDevice::Pause( void )
{
    m_Pause^=1;
    if(m_Pause){
        int i;
        for(i=0;i<4;i++){
            std::memset(&m_SndBuf[i],0,VCT_AUDIO_DATA_SIZE_MAX);
        }
    }
    return m_Pause;
}


/*---------------------------------------------------------------------------*
  Name        : FrameProcessThreadImpl
  Description : Frame Process Thread function.(Core)
  Arguments   : -
  Returns     : -
 *---------------------------------------------------------------------------*/
void SndMicDevice::FrameProcessThreadImpl( void )
{
    uptr Position;
    u32 micScale;
    u32 micOffset;
    nn::Result  Result;
    u16         outCIDList[VCT_MAX_AUDIO_STREAM];
//    char zero_buf[192*8*2];

    nn::os::Timer timer(false);

    m_SndBufWriteIdx  = 0;
    m_SndBufReadIdx   = 0;


    while ( true )
    {
        if ( m_MicStatus == MIC_PROCESS_NONE )
        {
            break;
        }

        /**********   MIC Start   **********/
        /* Check process status */
        if ( m_MicStatus != MIC_PROCESS_INIT ) 
        {
            SAMPLE_PRINT( "MIC_ERROR_SEQ !!\n" );
            break;
        }

        /* Wait for signal from sound thread */
        m_MicSyncEvent.Wait();

        /* Set Half Position */
        m_MicBuffHalfPosition = (uptr)m_MicBuffer + m_AudioFrameSize;
        m_MicPrevPosition     = 1;
    
        /* Start sampling. */
        Result = nn::mic::StartSampling(
                                            nn::mic::SAMPLING_TYPE_SIGNED_16BIT,
                                            nn::mic::SAMPLING_RATE_8180,
                                            0,
                                            m_AudioFrameSize*2,
                                            true
                                           );
    
        /* StartSampling error check. */
        if ( Result.IsFailure() )
        {
            /* Continue processing because control can reach this point with system closed     */
            SAMPLE_PRINT( "StartSampling error !!\n" );
            m_MicTimingError = true;
        }
        else
        {
            m_MicTimingError = false;
        }

        /* Update process status */
        m_MicStatus = MIC_PROCESS_START;


        timer.StartPeriodic(nn::fnd::TimeSpan::FromNanoSeconds(FREQ8KHZ_TIME_NS * 8 * m_Config.frame_msec),
            nn::fnd::TimeSpan::FromNanoSeconds(FREQ8KHZ_TIME_NS * 8 * m_Config.frame_msec));

        while (( m_MicStatus == MIC_PROCESS_START ) && ( m_SoundStatus == SND_PROCESS_PLAY ))
        {
            timer.Wait();        /* Launch for each sound frame */

            if( m_MicTimingError )
            {
                /* Restart mic sampling */
                /* Set Half Position */
                m_MicBuffHalfPosition = (uptr)m_MicBuffer + m_AudioFrameSize;
                m_MicPrevPosition     = 0;
            
                /* Start sampling. */
                Result = nn::mic::StartSampling(
                                                    nn::mic::SAMPLING_TYPE_SIGNED_16BIT,
                                                    nn::mic::SAMPLING_RATE_8180,
                                                    0,
                                                    m_AudioFrameSize*2,
                                                    true
                                                   );
            
                /* StartSampling error check. */
                if ( Result.IsFailure() )
                {
                    /* Continue processing because control can reach this point with system closed     */
                    SAMPLE_PRINT( "StartSampling error !!\n" );
                    m_MicTimingError = true;
                }
                else
                {
                    m_MicTimingError = false;
                }
            }

            /* Timer not synchronized with mic sampling, so keep some margin when getting mic */
            nn::os::Thread::Sleep( nn::fnd::TimeSpan::FromMilliSeconds( MIC_CAPTURE_MARGIN_MS ) );

            /* Get MIC Sampling Position. */
            if ( nn::mic::GetLastSamplingAddress( &Position ).IsFailure() )
            {
                SAMPLE_PRINT( "Cannot get Mic address!!\n" );
                //break;
            }

            /* Get Mic Data */
            if (m_MicPrevPosition == 1)
            {
                if ((((Position + 2) >= m_MicBuffHalfPosition) && ((Position + 2) < (m_MicBuffHalfPosition + m_AudioFrameSize)))
                       && ( m_MicTimingError == false))
                {
                    /* Handle MIC data. */
                    micOffset = static_cast<u32>(Position + 2 - m_MicBuffHalfPosition)/2;
                    micScale = VCT_GetAudioScale( (void *)m_MicBuffer, m_AudioFrameSize );
                    if(m_Pause == 0){
                        VCT_SendAudioEx( (void *)m_MicBuffer, micScale, micOffset );
                    }
                    m_MicPrevPosition = 0;
                }
                else
               {
                    /* Timer not synchronized with mic sampling, so control can reach this point */
                    /* Control also reaches this point when the system is closed     */
#if 1
                    SAMPLE_PRINT( "Wrong Mic Position 1 !!\n" );
#endif
                    /* Pass muted data */
                    micOffset = MIC_CAPTURE_MARGIN_SAMPLE;
                    std::memset( m_zero_buf, 0x0, m_AudioFrameSize );
                    micScale = 0;
                    if(m_Pause == 0){
                        VCT_SendAudioEx( m_zero_buf, micScale, micOffset );
                    }

                    /* Weird position, so restart mic sampling */
                    m_MicTimingError = true;
                }
            }
            else 
            {
                /* m_MicPrevPosition == 0  */
                if ((((Position + 2) < m_MicBuffHalfPosition) || ((Position + 2) == (m_MicBuffHalfPosition + m_AudioFrameSize)))
                       && ( m_MicTimingError == false))
                {
                    /* Handle MIC data. */
                    micOffset = static_cast<u32>(Position + 2 - reinterpret_cast<u32>(m_MicBuffer))/2;
                    if((Position + 2) == (m_MicBuffHalfPosition + m_AudioFrameSize))
                    {
                        micOffset = 0;
                    }
                    micScale = VCT_GetAudioScale( (void *)m_MicBuffHalfPosition, m_AudioFrameSize );
                    if(m_Pause == 0){
                        VCT_SendAudioEx( (void *)m_MicBuffHalfPosition, micScale, micOffset );
                    }
                    m_MicPrevPosition = 1;
                }
                else
                {
                    /* Timer not synchronized with mic sampling, so control can reach this point */
                    /* Control also reaches this point when the system is closed     */
#if 1
                    SAMPLE_PRINT( "Wrong Mic Position 0 !!\n" );
#endif
                    /* Pass muted data */
                    micOffset = MIC_CAPTURE_MARGIN_SAMPLE;
                    micScale = 0;
                    std::memset( m_zero_buf, 0x0, m_AudioFrameSize );
                    if(m_Pause == 0){
                        VCT_SendAudioEx( m_zero_buf, micScale, micOffset );
                    }

                    /* Weird position, so restart mic sampling */
                    m_MicTimingError = true;
                }
            }

            /* Set snd data. */
            if(m_Pause == 0){
                //N2050 20130329 
                VCT_ReceiveAudio( (void *)m_SndBuf[m_SndBufWriteIdx], m_AudioFrameSize, outCIDList );
            }

            m_SndBufWriteIdx = (m_SndBufWriteIdx + 1) % 4;
        }

        /* Stop sampling. */
        if ( nn::mic::StopSampling().IsFailure() )
        {
            SAMPLE_PRINT( "StopSampling Error!!\n" );
        }

        /* Notification when the microphone stops. */
        m_MicStopEvent.Signal();

        /* Wait until the status changes. */
        while (( m_MicStatus != MIC_PROCESS_NONE ) && ( m_SoundStatus != SND_PROCESS_INIT ))
        {
            timer.Wait();
        }
        timer.Stop();
        if( m_MicStatus == MIC_PROCESS_NONE )
        {
            break;
        }
    }

    timer.Finalize();

}


/*---------------------------------------------------------------------------*
  Name        : FrameProcessThreadFunc
  Description : Frame Process Thread function.(Wrapper)
  Arguments   : sndmicDev - SndMicDevice Class pointer.
  Returns     : -
 *---------------------------------------------------------------------------*/
void FrameProcessThreadFunc( SndMicDevice* sndmicDev )
{
    sndmicDev->FrameProcessThreadImpl();
}

#else       //#ifdef VCT_PLATFORM_CTR

/*---------------------------------------------------------------------------*
 * Name        : InitializeMic
 * Description : Initialize mic.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::InitializeMic( void )
{
    mic_ringbuffer_t micRb;
    int              result;

    // Check sequence.
    if ( m_micSeq != SNDMIC_SEQ_NONE )
    {
        SAMPLE_PRINT( "InitializeMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    if(m_AttatchUsbMic)
    {
        mem_res_query_t mrq;
        result = USBMICOpenQuery(m_UsbHandle, &mrq);
        if(result == USB_MIC_ERROR_NONE)
        {
            usb_mic_ringbuffer_t    mic_rb;
            mem_res_t               mrt;
            mrt.num_mem_res = 0;//mrq.num_mem_res;
            mrt.p_mem_res = NULL;//(mem_res_element_t *)MEMAllocFromDefaultHeapEx( mrq.num_mem_res * sizeof(mem_res_element_t), sizeof(long int) );
            mic_rb.modulus = MIC_RING_BUFFER_SIZE_BYTES / sizeof(usb_mic_pcm_sample_t);
            mic_rb.base = (usb_mic_pcm_sample_t*)MEMAllocFromDefaultHeapEx(MIC_RING_BUFFER_SIZE_BYTES, PPC_IO_BUFFER_ALIGN);
            result = USBMICOpen(m_UsbHandle,&mrt,&mic_rb);
            if(result == USB_MIC_ERROR_NONE)
            {
                m_micRingBuffer = mic_rb.base;
                m_UsbResBuffer = mrt.p_mem_res;

                SAMPLE_PRINT("USBMICOpen OK %X %X,%X\n",m_UsbHandle,mrt.num_mem_res,mic_rb.modulus);

                // Update sequence => Initialized.
                m_micSeq = MIC_SEQ_INITED;
                // Use usb mic.
                m_UseUsbMic = 1;
                return SNDMIC_ERROR_NONE;
            }
            else
            {
                SAMPLE_PRINT("USBMICOpen Error\n");
                MEMFreeToDefaultHeap( mrt.p_mem_res );
                MEMFreeToDefaultHeap( mic_rb.base );
                m_micRingBuffer = NULL;
            }
        }
        else
        {
            SAMPLE_PRINT("USBMICOpenQuery Error\n");
        }
    }

    // Use drc mic.
    m_UseUsbMic = 0;

    // Malloc mic buffer.
    m_micRingBuffer = (s16*)MEMAllocFromDefaultHeapEx( MIC_RING_BUFFER_SIZE_BYTES, PPC_IO_BUFFER_ALIGN );
    if ( m_micRingBuffer == NULL )
    {
        SAMPLE_PRINT( "InitializeMic Err [malloc memory].\n" );
        return MIC_ERROR_SYS;
    }

    // Set information of mic buffer.
    micRb.modulus = MIC_RING_BUFFER_SIZE_BYTES / sizeof( mic_pcm_sample_t );
    micRb.base    = ( mic_pcm_sample_t* )m_micRingBuffer;

    // Get mic handle.
    m_hMic = MICInit( mic_inst_0, NULL, &micRb, &result );

    if ( (int)m_hMic < MIC_ERROR_NONE )
    {
        // Free mic buffer.
        MEMFreeToDefaultHeap( m_micRingBuffer );
        m_micRingBuffer = NULL;

        SAMPLE_PRINT( "InitializeMic Err [MICInit(%d)].\n", result );
        return MIC_ERROR_SYS;
    }

    // Update sequence => Initialized.
    m_micSeq = MIC_SEQ_INITED;

    {//echo canceller
        u32 enable;
        int errcode;
        errcode = MICGetState(m_hMic,mic_state_echo_cancellation,&enable);
        if(errcode == 0 && enable == 0){
            MICSetState(m_hMic,mic_state_echo_cancellation,1);
            //MICGetState(m_hMic,mic_state_echo_cancellation,&enable);
            //SAMPLE_PRINT("SET ECHO=%d %d\n",enable,errcode);
        }
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : FinalizeMic
 * Description : Finalize mic.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::FinalizeMic( void )
{
    // Check sequence.
    if ( ( m_micSeq != MIC_SEQ_INITED ) && ( m_micSeq != MIC_SEQ_STOP ) )
    {
        SAMPLE_PRINT( "FinalizeMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Release mic handle.
    if(m_UseUsbMic == 0){
        MICUninit( m_hMic );
        m_hMic = -1;

        // Free mic buffer.
        if ( m_micRingBuffer != NULL )
        {
            MEMFreeToDefaultHeap( m_micRingBuffer );
            m_micRingBuffer = NULL;
        }
    }
    else{
        //USBMICUninit();
        // Free mic buffer.
        if ( m_micRingBuffer != NULL )
        {
            MEMFreeToDefaultHeap( m_micRingBuffer );
            m_micRingBuffer = NULL;
        }
        if ( m_UsbResBuffer != NULL )
        {
            MEMFreeToDefaultHeap( m_UsbResBuffer );
            m_UsbResBuffer = NULL;
        }
    }

    // Update sequence => None.
    m_micSeq = SNDMIC_SEQ_NONE;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StartMic
 * Description : Start mic.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StartMic( void )
{
    int             result;

    // Check sequence.
    if ( ( m_micSeq != MIC_SEQ_INITED ) && ( m_micSeq != MIC_SEQ_STOP ) )
    {
        SAMPLE_PRINT( "StartMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Start mic.
    if(m_UseUsbMic == 0){
        mic_status_t    micStatus;
        result = MICOpen( m_hMic );
        if ( result != MIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "StartMic Err [MICOpen(%d)].\n", result );
            return MIC_ERROR_SYS;
        }

        // Wait for mic ready.
        while ( true )
        {
            // Get mic status
            result = MICGetStatus( m_hMic, &micStatus );
            if ( result != MIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetStatus(%d)].\n", result );
                return MIC_ERROR_SYS;
            }

            // Check available mic data.
            if ( micStatus.available > ( MIC_RING_BUFFER_SIZE_BYTES >> 2 ) )
            {
                break;
            }

            OSSleepMilliseconds( 3 );
        }

        // Initialize mic gain.
        while ( 1 )
        {
            OSSleepMilliseconds( 3 );
            result = MICGetState( m_hMic, mic_state_sample_rate, &m_micSampleRate );
            if ( result != SNDMIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState mic_state_sample_rate(%d)].\n", result );
//                return MIC_ERROR_SYS;
                continue;
            }

            // Get gain step.
            result = MICGetState( m_hMic, mic_state_gain_step, &m_micGainStep );
            if ( result != SNDMIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState mic_state_gain_step(%d)].\n", result );
//                return MIC_ERROR_SYS;
                continue;
            }

            // Get gain min.
            result = MICGetState( m_hMic, mic_state_gain_min, &m_micGainMin );
            if ( result != SNDMIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState mic_state_gain_min(%d)].\n", result );
//                return MIC_ERROR_SYS;
                continue;
            }

            // Get gain max.
            result = MICGetState( m_hMic, mic_state_gain_max, &m_micGainMax );
            if ( result != SNDMIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState mic_state_gain_max(%d)].\n", result );
//                return MIC_ERROR_SYS;
                continue;
            }

            // Get gain.
            result = MICGetState( m_hMic, mic_state_gain_db, &m_micGainLevel );
            if ( result != SNDMIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState mic_state_gain_db(%d)].\n", result );
//                return MIC_ERROR_SYS;
                continue;
            }

            if (( m_micGainStep  & 0xFFFF8000 ) ||
                ( m_micGainMin   & 0xFFFF8000 ) ||
                ( m_micGainMax   & 0xFFFF8000 ) ||
                ( m_micGainLevel & 0xFFFF8000 ))
            {
                SAMPLE_PRINT( "StartMic Err [MICGetState].\n" );
                return MIC_ERROR_SYS;
            }

            // Set Default mic gain.
            m_micGainLevel = m_micGainMax;
            result = MICSetState( m_hMic, mic_state_gain_db, m_micGainMax );
            if ( result != 0x0 )
            {
                return MIC_ERROR_SYS;
            }
            
            break;  // Escape from while loop
        }  // end of while
    }
    else
    {
        usb_mic_status_t    micStatus;
        SAMPLE_PRINT("USBMICStart\n");
        result = USBMICStart(m_UsbHandle);
        if(result == USB_MIC_ERROR_NONE)
        {
            // Wait for mic ready.
            while ( true )
            {
                // Get mic status
                result = USBMICGetStatus(m_UsbHandle, &micStatus);
                if ( result != MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [USBMICGetStatus(%d)].\n", result );
                    return MIC_ERROR_SYS;
                }

                // Check available mic data.
                if ( micStatus.available > ( MIC_RING_BUFFER_SIZE_BYTES >> 2 ) )
                {
                    break;
                }

                OSSleepMilliseconds( 3 );
            }

            // Initialize mic gain.
            while ( 1 )
            {
                OSSleepMilliseconds( 3 );
                result = USBMICGetState( m_UsbHandle, usb_mic_state_sample_rate, &m_micSampleRate );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [MICGetState usb_mic_state_sample_rate(%d)].\n", result );
//                    return MIC_ERROR_SYS;
                    continue;
                }

                // Get gain step.
                result = USBMICGetState( m_UsbHandle, usb_mic_state_gain_step, &m_micGainStep );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [MICGetState usb_mic_state_gain_step(%d)].\n", result );
//                    return MIC_ERROR_SYS;
                    continue;
                }

                // Get gain min.
                result = USBMICGetState( m_UsbHandle, usb_mic_state_gain_min, &m_micGainMin );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [MICGetState usb_mic_state_gain_min(%d)].\n", result );
//                    return MIC_ERROR_SYS;
                    continue;
                }

                // Get gain max.
                result = USBMICGetState( m_UsbHandle, usb_mic_state_gain_max, &m_micGainMax );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [MICGetState usb_mic_state_gain_max(%d)].\n", result );
//                    return MIC_ERROR_SYS;
                    continue;
                }

                // Get gain.
                result = USBMICGetState( m_UsbHandle, usb_mic_state_gain_db, &m_micGainLevel );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    SAMPLE_PRINT( "StartMic Err [MICGetState usb_mic_state_gain_db(%d)].\n", result );
//                    return MIC_ERROR_SYS;
                    continue;
                }

                // Set Default mic gain.
                m_micGainLevel = m_micGainMax;
                result = USBMICSetState( m_UsbHandle, usb_mic_state_gain_db, m_micGainMax );
                if ( result != USB_MIC_ERROR_NONE )
                {
                    return MIC_ERROR_SYS;
                }

                break;  // Escape from while loop
            }  // end of while
        }
        else
        {
            SAMPLE_PRINT("USBMICStart Error\n");
            return MIC_ERROR_SYS;
        }
    }

    // Update sequence => Start.
    m_micSeq = MIC_SEQ_START;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StopMic
 * Description : Stop mic.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StopMic( void )
{
    int result;

    // Check sequence.
    if ( m_micSeq != MIC_SEQ_START )
    {
        SAMPLE_PRINT( "StopMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Stop mic.
    if(m_UseUsbMic == 0){
        result = MICClose( m_hMic );
        if ( result != 0 )
        {
            SAMPLE_PRINT( "StopMic Err [MICClose(%d)].\n", result );
            return MIC_ERROR_SYS;
        }
    }
    else{
        USBMICStop(m_UsbHandle);
        result = USBMICClose(m_UsbHandle);
        if ( result != USB_MIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "StopMic Err [MICClose(%d)].\n", result );
            return MIC_ERROR_SYS;
        }
    }

    // Update sequence => Stop.
    m_micSeq = MIC_SEQ_STOP;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : ReadMic
 * Description : Read mic data.
 * Arguments   : out_mic_data - Buffer of mic data.
 *             : in_sample    - Request of mic data sample.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_WRN_FEW_DATA - None error, but few data.
 *             :    MIC_ERROR_PARAM - Parameter error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::ReadMic( s16 *out_micData, s16 in_sample )
{
    s16             *dstBuf;
    int             readIdx;
    int             result;
    int             cnt;

    // Check sequence.
    if ( m_micSeq != MIC_SEQ_START )
    {
        SAMPLE_PRINT( "ReadMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Check parameter.
    if ( ( out_micData == NULL ) || ( in_sample <= 0 ) )
    {
        SAMPLE_PRINT( "ReadMic Err [parameter : 1st arg(%lu), 2nd arg(%d)].\n", reinterpret_cast<u32>(out_micData), in_sample );
        return MIC_ERROR_PARAM;
    }

    // Get mic status
    if(m_UseUsbMic == 0)
    {
        mic_status_t    micStatus;
        result = MICGetStatus( m_hMic, &micStatus );
        if ( result != MIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "ReadMic Err [MICGetStatus(%d)].\n", result );
            return MIC_ERROR_SYS;
        }

        // Check available mic data.
        if ( micStatus.available < in_sample )
        {
            SAMPLE_PRINT( "ReadMic Wrn [few data(%d)].\n", micStatus.available );
            return MIC_WARN_FEW_DATA;
        }

        // Read mic data.
        readIdx = micStatus.read_idx;
        dstBuf  = out_micData;
        for ( cnt = 0; cnt < in_sample; cnt++ )
        {
            // Copy data.
            *dstBuf = m_micRingBuffer[readIdx];

            // Update index.
            dstBuf++;
            readIdx++;
            if ( readIdx >= ( MIC_RING_BUFFER_SIZE_BYTES / sizeof( mic_pcm_sample_t ) ) )
            {
                readIdx -= MIC_RING_BUFFER_SIZE_BYTES / sizeof( mic_pcm_sample_t );
            }
        }

        // Consume.
        result = MICSetDataConsumed( m_hMic, in_sample );
        if ( result != 0x0 )
        {
            SAMPLE_PRINT( "ReadMic Err [MICSetDataConsumed(%d)].\n", result );
            return MIC_ERROR_SYS;
        }
    }
    else
    {
        usb_mic_status_t    micStatus;
        result = USBMICGetStatus(m_UsbHandle, &micStatus);
        if ( result != USB_MIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "ReadMic Err [MICGetStatus(%d)].\n", result );
            return MIC_ERROR_SYS;
        }

        // Check available mic data.
        if ( micStatus.available < in_sample )
        {
            SAMPLE_PRINT( "ReadMic Wrn [few data(%d)].\n", micStatus.available );
            return MIC_WARN_FEW_DATA;
        }

        // Read mic data.
        readIdx = micStatus.read_idx;
        dstBuf  = out_micData;
        for ( cnt = 0; cnt < in_sample; cnt++ )
        {
            // Copy data.
            *dstBuf = m_micRingBuffer[readIdx];

            // Update index.
            dstBuf++;
            readIdx++;
            if ( readIdx >= ( MIC_RING_BUFFER_SIZE_BYTES / sizeof( mic_pcm_sample_t ) ) )
            {
                readIdx -= MIC_RING_BUFFER_SIZE_BYTES / sizeof( mic_pcm_sample_t );
            }
        }

        // Consume.
        result = USBMICSetDataConsumed( m_UsbHandle, in_sample );
        if ( result != USB_MIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "ReadMic Err [MICSetDataConsumed(%d)].\n", result );
            return MIC_ERROR_SYS;
        }
    }
    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : GetStateMic
 * Description : Get mic state.
 * Arguments   : *out_state - state of mic.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_PARAM - Parameter error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::GetStateMic( MIC_STATE_INFO *out_state ) const
{
    // Check sequence.
    if ( m_micSeq != MIC_SEQ_START )
    {
        SAMPLE_PRINT( "GetStateMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Check parameter.
    if ( out_state == NULL )
    {
        SAMPLE_PRINT( "GetStateMic Err [parameter : arg(%lu)].\n", out_state );
        return MIC_ERROR_PARAM;
    }

    // Teach mic state.
    out_state->sampleRate     = m_micSampleRate;
    out_state->micGainMin     = m_micGainMin;
    out_state->micGainMax     = m_micGainMax;
    out_state->micGainStep    = m_micGainStep;
    out_state->micGainLevel   = m_micGainLevel;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : SetGainLevelMic
 * Description : Set mic gain level.
 * Arguments   : in_gainLevel - value of set mic gain level(0-100).
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_PARAM - Parameter error.
 *             :    MIC_ERROR_SYS   - System error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetGainLevelMic( unsigned int in_gainLevel )
{
    unsigned int val=~0;
    int          result;

    // Check sequence.
    if ( m_micSeq != MIC_SEQ_START )
    {
        SAMPLE_PRINT( "SetGainMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Check parameter.
    if (( in_gainLevel > m_micGainMax ) || ( in_gainLevel < m_micGainMin ))
    {
        SAMPLE_PRINT( "SetGainMic Err [parameter(%d)].\n", in_gainLevel );
        return MIC_ERROR_PARAM;
    }

    // Set mic gain.
    if ( in_gainLevel != m_micGainLevel )
    {
        // Check mic state.
        if(m_UseUsbMic == 0)
        {
            MICGetState( m_hMic, mic_state_gain_db, &val );
            if ( val & 0xFFFF8000 )
            {
                SAMPLE_PRINT( "SetGainMic Err [MICGetState mic_state_gain_db(%d)].\n", val );
                return MIC_ERROR_SYS;
            }

            // Set mic gain.
            result = MICSetState( m_hMic, mic_state_gain_db, in_gainLevel );
            if ( result != 0x0 )
            {
                SAMPLE_PRINT( "SetGainMic Err [MICSetState mic_state_gain_db(%d)].\n", result );
                return MIC_ERROR_SYS;
            }
        }
        else
        {
            USBMICGetState( m_UsbHandle, usb_mic_state_gain_db, &val );

            // Set mic gain.
            result = USBMICSetState( m_UsbHandle, usb_mic_state_gain_db, in_gainLevel );
            if ( result != USB_MIC_ERROR_NONE )
            {
                SAMPLE_PRINT( "SetGainMic Err [MICSetState usb_mic_state_gain_db(%d)].\n", result );
                return MIC_ERROR_SYS;
            }
        }

        // Update current mic gain level.
        m_micGainLevel = in_gainLevel;
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : GetGainLevelMic
 * Description : Get mic gain level.
 * Arguments   : *out_gainLevel - value of get mic gain level(0-100).
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    MIC_ERROR_SEQ  - Sequence error.
 *             :    MIC_ERROR_PARAM - Parameter error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::GetGainLevelMic( unsigned int *out_gainLevel ) const
{
    // Check sequence.
    if ( m_micSeq != MIC_SEQ_START )
    {
        SAMPLE_PRINT( "GetGainLevelMic Err [sequence(%d)].\n", m_micSeq );
        return MIC_ERROR_SEQ;
    }

    // Check parameter.
    if ( out_gainLevel == NULL )
    {
        SAMPLE_PRINT( "GetGainLevelMic Err [parameter : arg(%lu)].\n", out_gainLevel );
        return MIC_ERROR_PARAM;
    }

    // Teach mic gain level.
    *out_gainLevel = m_micGainLevel;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : InitializeSnd
 * Description : Initialize sound.
 * Arguments   : *in_sndParam - sound parameter.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SND_ERROR_NONE  - None error.
 *             :    SND_ERROR_PARAM - Parameter error.
 *             :    SND_ERROR_SYS   - System error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::InitializeSnd( const SND_PARAM *in_sndParam )
{
    AXPBOFFSET      addr;
    f32             ratio;
    s32             busyLength;
    s32             noiseLength;
    FSClient hFSClient;
    FSCmdBlock hFSCmdBlock;

    // Check sequence.
    if ( m_sndSeq != SNDMIC_SEQ_NONE )
    {
        SAMPLE_PRINT( "InitializeSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Check parameter.
    if ( in_sndParam == NULL )
    {
        SAMPLE_PRINT( "InitializeSnd Err [parameter : arg(%lu)].\n", reinterpret_cast<u32>(in_sndParam) );
        return SND_ERROR_PARAM;
    }
    if ( in_sndParam->freq >= SND_FREQ_MAX )
    {
        SAMPLE_PRINT( "InitializeSnd Err [parameter : member].\n" );
        return SND_ERROR_PARAM;
    }

    //Initialize FileSystem
    FSInit();
    FSAddClient(&hFSClient,FS_RET_NO_ERROR);
    FSInitCmdBlock(&hFSCmdBlock);

    // Open busy data.
    busyLength = readFileData(BUSY_FILE_PATH,&hFSClient,&hFSCmdBlock,&m_busyBuffer);

    // Open noise data.
    noiseLength = readFileData(NOISE_FILE_PATH,&hFSClient,&hFSCmdBlock,&m_noiseBuffer);

    //Shutdown FileSystem
    FSDelClient(&hFSClient,FS_RET_NO_ERROR);
    FSShutdown();
    if(busyLength < 0 || noiseLength < 0)
    {
        return SND_ERROR_SYS;
    }

    // Store parameter.
    m_freq             = in_sndParam->freq;

    // Calc sample per frame.
    if ( m_freq == SND_FREQ_08KHZ )
    {
        m_frameSample = in_sndParam->frameSample * 8;
    }
    else if ( m_freq == SND_FREQ_16KHZ )
    {
        m_frameSample = in_sndParam->frameSample * 16;
    }
    else // if ( m_freq == SND_FREQ_32KHZ )
    {
        m_frameSample = in_sndParam->frameSample * 32;
    }

    // Malloc sound buffer.
    m_streamBuffer = NULL;
    m_streamBuffer = (s16*)MEMAllocFromDefaultHeapEx( STREAMBUFFER_SIZE, PPC_IO_BUFFER_ALIGN );
    if ( m_streamBuffer == NULL )
    {
        SAMPLE_PRINT( "InitializeSnd Err [malloc stream buffer].\n" );
        return SND_ERROR_SYS;
    }
    memset( m_streamBuffer, 0x0, STREAMBUFFER_SIZE );

    // Malloc VoiceChat data buffer.
    m_voiceBuffer = NULL;
    m_voiceBuffer = (s16*)MEMAllocFromDefaultHeapEx( VOICEBUFFER_BASE_SIZE * m_frameSample, PPC_IO_BUFFER_ALIGN );
    if ( m_voiceBuffer == NULL )
    {
        SAMPLE_PRINT( "InitializeSnd Err [malloc voicechat data buffer].\n" );
        return SND_ERROR_SYS;
    }
    memset( m_voiceBuffer, 0x0, VOICEBUFFER_BASE_SIZE * m_frameSample );

    // Initialize voice buffer control variable.
    m_voiceWriteIdx    = 0x0;
    m_voiceReadIdx     = 0x0;
    m_voiceAvailable   = 0x0;
    m_prevPtr          = 0x0;
    m_totalSample      = 0x0;

    // Initialize AX.
    AXInit();
    AXSetMode( AX_MODE_STEREO );
    AXSetDefaultMixerSelect( AX_PB_MIXER_SELECT_DSP );

    // Initialize MIX.
    MIXInit();
    MIXSetSoundMode( MIX_SOUND_MODE_STEREO );

    // Register audio frame callback.
    AXRegisterFrameCallback( &axFrameCallback );

    // <----
    AXUserBegin();

    // Acquire voice
    m_voice = AXAcquireVoice( 15, NULL, 0 );
    if ( m_voice == NULL )
    {
        // Free sound buffer.
        MEMFreeToDefaultHeap( m_streamBuffer );
        m_streamBuffer = NULL;

        // Free VoiceChat data buffer.
        MEMFreeToDefaultHeap( m_voiceBuffer );
        m_voiceBuffer = NULL;

        SAMPLE_PRINT( "InitializeSnd Err [AXAcquireVoice].\n" );
        return SND_ERROR_SYS;
    }

    // Set mix channel.
    MIXInitChannel( m_voice, 0, 0, -904, -904, -904, 64, 127, 0 );
    MIXDRCInitChannel( m_voice, 0, 0, 64, 0 );
    MIXUpdateSettings();

    // Set voice offset.
    addr.loopFlag      = AXPBADDR_LOOP_ON;
    addr.format        = AX_PB_FORMAT_PCM16;
    addr.samples       = m_streamBuffer;
    addr.loopOffset    = 0x0;
    addr.endOffset     = ( STREAMBUFFER_SIZE >> 1 ) - 1;
    addr.currentOffset = 0x0;
    AXSetVoiceOffsets( m_voice, &addr );

    // Set voice src.
    AXSetVoiceSrcType( m_voice, AX_SRC_TYPE_LINEAR );
    if ( m_freq == SND_FREQ_08KHZ )
    {
        ratio      = 0.25f;
        m_writeIdx = STREAMBUFFER_BLOCKS * 2;
        m_maxIdx   = STREAMBUFFER_BLOCKS * 4;
        m_blkSize  = (AX_IN_SAMPLES_PER_FRAME) >> 2;
    }
    else if ( m_freq == SND_FREQ_16KHZ )
    {
        ratio      = 0.50f;
        m_writeIdx = STREAMBUFFER_BLOCKS * 1;
        m_maxIdx   = STREAMBUFFER_BLOCKS * 2;
        m_blkSize  = (AX_IN_SAMPLES_PER_FRAME) >> 1;
    }
    else // if ( m_freq == SND_FREQ_32KHZ )
    {
        ratio      = 1.00f;
        m_writeIdx = ( STREAMBUFFER_BLOCKS >> 1 );
        m_maxIdx   = STREAMBUFFER_BLOCKS;
        m_blkSize  = AX_IN_SAMPLES_PER_FRAME;
    }
    AXSetVoiceSrcRatio( m_voice, ratio );

    // Prepare data of stream buffer.
    AXPrepareEfxData( m_streamBuffer, STREAMBUFFER_SIZE );

// For busy/noise.
    m_busy = AXAcquireVoice( 15, NULL, 0 );
    MIXInitChannel( m_busy, 0, 0, -904, -904, -904, 64, 127, 0 );
    MIXDRCInitChannel( m_busy, 0, 0, 64, 0 );
    MIXUpdateSettings();
    addr.loopFlag      = AXPBADDR_LOOP_OFF;
    addr.format        = AX_PB_FORMAT_PCM16;
    addr.samples       = m_busyBuffer;
    addr.loopOffset    = 0x0;
    addr.endOffset     = ( busyLength >> 1 ) - 1;
    addr.currentOffset = 0x0;
    AXSetVoiceOffsets( m_busy, &addr );
    AXSetVoiceSrcType( m_busy, AX_SRC_TYPE_LINEAR );
    AXSetVoiceSrcRatio( m_busy, 1.0f );
    AXPrepareEfxData( m_busyBuffer, busyLength );

    m_noise = AXAcquireVoice( 15, NULL, 0 );
    MIXInitChannel( m_noise, 0, 0, -904, -904, -904, 64, 127, 0 );
    MIXDRCInitChannel( m_noise, 0, 0, 64, 0 );
    MIXUpdateSettings();
    addr.loopFlag      = AXPBADDR_LOOP_OFF;
    addr.format        = AX_PB_FORMAT_PCM16;
    addr.samples       = m_noiseBuffer;
    addr.loopOffset    = 0x0;
    addr.endOffset     = ( noiseLength >> 1 ) - 1;
    addr.currentOffset = 0x0;
    AXSetVoiceOffsets( m_noise, &addr );
    AXSetVoiceSrcType( m_noise, AX_SRC_TYPE_LINEAR );
    AXSetVoiceSrcRatio( m_noise, 1.0f );
    AXPrepareEfxData( m_noiseBuffer, noiseLength );

    //   ---->
    AXUserEnd();

    // Update sequence => Initialized.
    m_sndSeq = SND_SEQ_INITED;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : FinalizeSnd
 * Description : Finalize sound.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::FinalizeSnd( void )
{
    // Check sequence.
    if ( ( m_sndSeq != SND_SEQ_INITED ) && ( m_sndSeq != SND_SEQ_STOPPING ) && ( m_sndSeq != SND_SEQ_STOPPED ) )
    {
        SAMPLE_PRINT( "FinalizeSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Wait stopped.
    while ( ( m_sndSeq != SND_SEQ_INITED ) && ( m_sndSeq != SND_SEQ_STOPPED ) )
    {
        // No work.
        OSSleepMilliseconds( 1 );
    }

    // Finalize voice.
    MIXReleaseChannel( m_voice );
    AXFreeVoice( m_voice );
    m_voice = NULL;

    // Shutdown MIX/AX.
    MIXQuit();
    AXQuit();

    // Free VoiceChat data buffer.
    if ( m_voiceBuffer != NULL )
    {
        MEMFreeToDefaultHeap( m_voiceBuffer );
        m_voiceBuffer = NULL;
    }

    // Free sound buffer.
    if ( m_streamBuffer != NULL )
    {
        MEMFreeToDefaultHeap( m_streamBuffer );
        m_streamBuffer = NULL;
    }

    // Free noise buffer.
    if ( m_noiseBuffer != NULL )
    {
        MEMFreeToDefaultHeap( m_noiseBuffer );
        m_noiseBuffer = NULL;
    }

    // Free busy buffer.
    if ( m_busyBuffer != NULL )
    {
        MEMFreeToDefaultHeap( m_busyBuffer );
        m_busyBuffer = NULL;
    }

    // Update sequence => None.
    m_sndSeq = SNDMIC_SEQ_NONE;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StartSnd
 * Description : Start sound.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StartSnd( void )
{
    // Check sequence.
    if ( ( m_sndSeq != SND_SEQ_INITED ) && ( m_sndSeq != SND_SEQ_STOPPING ) && ( m_sndSeq != SND_SEQ_STOPPED ) )
    {
        SAMPLE_PRINT( "StartSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Start AX.
    AXSetVoiceState( m_voice, AX_PB_STATE_RUN );

    // Update sequence => Start.
    m_sndSeq = SND_SEQ_START;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StopSnd
 * Description : Stop sound.
 * Arguments   : None.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StopSnd( void )
{
    // Check sequence.
    if ( m_sndSeq != SND_SEQ_START )
    {
        SAMPLE_PRINT( "StopSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Stop AX.
    AXSetVoiceState( m_voice, AX_PB_STATE_STOP );

    // Update sequence => Stopping.
    m_sndSeq = SND_SEQ_STOPPING;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StartEffect
 * Description : Start effect sound.
 * Arguments   : in_effect - Effect sound number.(0 - noise, 1 - busy)
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StartEffect( u16 in_effect )
{
    // Start AX.
    if ( in_effect == 0 )
    {
        AXSetVoiceCurrentOffset( m_noise, 0x0 );
        AXSetVoiceState( m_noise, AX_PB_STATE_RUN );
    }
    else if ( in_effect == 1 )
    {
        AXSetVoiceCurrentOffset( m_busy, 0x0 );
        AXSetVoiceState( m_busy, AX_PB_STATE_RUN );
    }

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : StoreDataSnd
 * Description : Store voice data to snd.
 * Arguments   : *in_buffer - Store data buffer.
 *             : in_sample  - Store data sample.
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_SYS   - System error.
 *             :    SND_ERROR_PARAM - Parameter error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::StoreDataSnd( const s16 *in_buffer, s16 in_sample )
{
    // Check sequence.
    if ( m_sndSeq != SND_SEQ_START )
    {
        SAMPLE_PRINT( "StoreDataSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Check parameter.
    if ( ( in_buffer == NULL ) || ( in_sample != m_frameSample ) )
    {
        SAMPLE_PRINT( "StoreDataSnd Err [parameter 1st arg(%lu) 2nd arg(%d)].\n", reinterpret_cast<u32>(in_buffer), in_sample );
        return SND_ERROR_PARAM;
    }

    // Check overflow.
    if ( ( m_voiceAvailable + in_sample ) >= ( VOICEBUFFER_BLOCKS * m_frameSample ) )
    {
        return SND_ERROR_PARAM;
    }

    // Store voice data to voice buffer.
    memcpy( &m_voiceBuffer[m_voiceWriteIdx], in_buffer, in_sample * sizeof( s16 ) );

    // Update control.
    m_voiceWriteIdx   = ( m_voiceWriteIdx + in_sample ) % ( VOICEBUFFER_BLOCKS * m_frameSample );
    m_voiceAvailable += in_sample;

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : GetDRCGainSnd
 * Description : Get drc master gain.
 * Arguments   : *out_drcGain - Drc master volume.(value of gain, q15 format)
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_PARAM - Parameter error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::GetDRCGainSnd( u16 *out_drcGain ) const
{
    // Check sequence.
    if ( m_sndSeq == SNDMIC_SEQ_NONE )
    {
        SAMPLE_PRINT( "GetDRCGainSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Check parameter.
    if ( out_drcGain == NULL )
    {
        SAMPLE_PRINT( "GetDRCGainSnd Err [param : arg (%lu)].\n", reinterpret_cast<u32>(out_drcGain) );
        return SND_ERROR_PARAM;
    }

    // Get drc master volume.
    AXGetDeviceVolume( AX_DEVICE_DRC, AX_DRC_ID0, out_drcGain );

    return SNDMIC_ERROR_NONE;
}


/*---------------------------------------------------------------------------*
 * Name        : SetDRCGainSnd
 * Description : Set drc master gain.
 * Arguments   : in_drcGain - Drc master volume.(value of gain, q15 format)
 * Returns     : SNDMIC_ERROR_STATUS
 *             :    SNDMIC_ERROR_NONE - None error.
 *             :    SND_ERROR_PARAM - Parameter error.
 *             :    SND_ERROR_SEQ  - Sequence error.
 *---------------------------------------------------------------------------*/
SNDMIC_ERROR_STATUS SndMicDevice::SetDRCGainSnd( u16 in_drcGain)
{
    // Check sequence.
    if ( m_sndSeq == SNDMIC_SEQ_NONE )
    {
        SAMPLE_PRINT( "SetDRCGainSnd Err [sequence(%d)].\n", m_sndSeq );
        return SND_ERROR_SEQ;
    }

    // Check parameter.
    if ( in_drcGain > 0x8000 )
    {
        SAMPLE_PRINT( "SetDRCGainSnd Err [param : arg (%d)].\n", static_cast<u32>(in_drcGain) );
        return SND_ERROR_PARAM;
    }

    // Set drc master volume.
    AXSetDeviceVolume( AX_DEVICE_DRC, AX_DRC_ID0, in_drcGain );

    return SNDMIC_ERROR_NONE;
}


SNDMIC_ERROR_STATUS SndMicDevice::SetDRCVolume( u32 index )
{
    return SetDRCGainSnd( speakerGainTbl[index] );
}

/*---------------------------------------------------------------------------*
 * Internal Functions
 *---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*
 * Name        : axFrameCallback
 * Description : Callback every 3 msec by AX handler.
 * Arguments   : None.
 * Returns     : None.
 *---------------------------------------------------------------------------*/
void SndMicDevice::axFrameCallback( void )
{
    m_This->axFrameCallbackImpl();
}

void SndMicDevice::axFrameCallbackImpl( void )
{
    AXPBOFFSET  addr;
    u32         currPtr;
    u32         currIdx;

    // tell the mixer to update settings.
    MIXUpdateSettings();

    switch ( m_sndSeq )
    {
        case SND_SEQ_START :
            // Get voice offset information.
            AXGetVoiceOffsets( m_voice, &addr );

            // Calc stream buffer block current index.
            currPtr = addr.currentOffset;
            currIdx = currPtr  / m_blkSize;

            // Supplies 3sec data.
            if ( m_writeIdx != currIdx )
            {
                axSetData();
            }
            else
            {
                // Don't come here.
                SAMPLE_PRINT( "axFrameCallback Err [writeIdx == currIdx]\n" );
            }

            // Calc used data..
            if ( m_prevPtr > currPtr )
            {
                m_totalSample += ( ( STREAMBUFFER_SIZE >> 1) - ( m_prevPtr - currPtr ) );
            }
            else
            {
                m_totalSample += ( currPtr - m_prevPtr );
            }
            m_prevPtr = currPtr;

            // Request store data.
            if ( m_totalSample >= m_frameSample )
            {
                m_totalSample -= m_frameSample;
                OSSignalEvent( &m_AudioFrameEvent );
            }
            break;

        case SND_SEQ_STOPPING :
            m_sndSeq = SND_SEQ_STOPPED;
            break;

        case SNDMIC_SEQ_NONE :
        case SND_SEQ_INITED :
        case SND_SEQ_STOPPED :
        default:
            break;
    }
}


/*---------------------------------------------------------------------------*
 * Name        : axSetData
 * Description : Supplies 3sec data to AX voice.
 * Arguments   : None.
 * Returns     : None.
 *---------------------------------------------------------------------------*/
void SndMicDevice::axSetData( void )
{
    s16 tmpData[AX_IN_SAMPLES_PER_FRAME];

    // Read voice data from voice buffer.
    readVoiceData( tmpData, m_blkSize );

    // Set voice data.
    memcpy( &m_streamBuffer[m_blkSize * m_writeIdx], tmpData, sizeof(s16) * m_blkSize );
    AXPrepareEfxData( &m_streamBuffer[m_blkSize * m_writeIdx], sizeof(s16) * m_blkSize );

    // Update write index.
    m_writeIdx = ( m_writeIdx + 1 ) % m_maxIdx;
}


/*---------------------------------------------------------------------------*
 * Name        : readVoiceData
 * Description : Read voice data from voice buffer.
 * Arguments   : None.
 * Returns     : None.
 *---------------------------------------------------------------------------*/
void SndMicDevice::readVoiceData( s16 *out_data, s16 in_sample )
{
    // Check available.
    if ( m_voiceAvailable < in_sample )
    {
        // Set 0x0 data.
        memset( out_data, 0x0, in_sample * sizeof( s16 ) );
    }
    else
    {
        // Read voice data.
        memcpy( out_data, &m_voiceBuffer[m_voiceReadIdx], in_sample * sizeof( s16 ) );

        // Update control.
        m_voiceReadIdx    = ( m_voiceReadIdx + in_sample ) % ( VOICEBUFFER_BLOCKS * m_frameSample );
        m_voiceAvailable -= in_sample;
    }
}

s32 SndMicDevice::readFileData(const char *filePath,FSClient *hFSClient,FSCmdBlock *hFSCmdBlock,s16 **buffer)
{
    FSFileHandle fh;
    FSStat stat;
    FSStatus result;
    char openPath[256];
    s32 length;

    // Open data.
    strcpy(openPath,"/vol/content/");
    strcat(openPath,filePath);
    result = FSOpenFile(hFSClient,hFSCmdBlock,openPath,"r",&fh,FS_RET_NO_ERROR);
    if(result == FS_STATUS_OK){
        // Get length.
        FSGetStatFile(hFSClient,hFSCmdBlock,fh,&stat,FS_RET_NO_ERROR);
        length = stat.size;
        if ( length != -1){
            *buffer = (s16*)MEMAllocFromDefaultHeapEx( length, PPC_IO_BUFFER_ALIGN );
            if ( *buffer != NULL )
            {
                // Read busy data.
                result = FSReadFile(hFSClient,hFSCmdBlock,*buffer,static_cast<u32>(length),1,fh,0,FS_RET_NO_ERROR);
                if(result < 0)
                {
                    length = (-1);
                    SAMPLE_PRINT( "InitializeSnd Err [Read data].\n" );
                }
            }
            else{
                length = (-1);
                SAMPLE_PRINT( "InitializeSnd Err [malloc buffer].\n" );
            }
        }
        else{
            SAMPLE_PRINT( "InitializeSnd Err [can't get length audio file].\n" );
        }
        // Close busy data.
        result = FSCloseFile(hFSClient,hFSCmdBlock,fh,FS_RET_NO_ERROR);
        if( result != FS_STATUS_OK )
        {
            length = (-1);
            SAMPLE_PRINT( "InitializeSnd Err [Close audio file].\n" );
        }
    }
    else{
        length = (-1);
        SAMPLE_PRINT( "InitializeSnd Err [can't open audio file].\n" );
    }

    return length;
}

static int MicSndFrameThread( int,void *p )
{
    SndMicDevice *This = reinterpret_cast<SndMicDevice *>(p);
    return This->MicSndFrameThreadImpl();
}

/*---------------------------------------------------------------------------*
  Name:         RSMP32to8

  Description:  Downsamples from 32 kHz to 8 kHz.
                The size of the output array is the size of the input divided by 4.

  Arguments:    input        input sample
                num          number of input samples
                output       output sample

  Returns:      s32          number of output samples
 *---------------------------------------------------------------------------*/
s32 SndMicDevice::RSMP32to8(s16 *input, s32 num, s16* output)
{
    s32  ii, jj;

    for(ii = jj = 0 ; ii < num ; ii+=4){
        output[jj++] = input[ii];
    }
    return jj;
}

/*---------------------------------------------------------------------------*
  Name:         RSMP32to16

  Description:  Downsamples from 32 kHz to 16 kHz.
                The size of the output array is the size of the input divided by 2.

  Arguments:    input        input sample
                num          number of input samples
                output       output sample

  Returns:      s32          number of output samples
 *---------------------------------------------------------------------------*/
s32 SndMicDevice::RSMP32to16(s16 *input, s32 num, s16* output)
{
    s32  ii, jj;

    for(ii = jj = 0 ; ii < num ; ii+=2){
        output[jj++] = input[ii];
    }
    return jj;
}

/*---------------------------------------------------------------------------*
  Name        : Pause
  Description : pause or restart
  Arguments   : input       SND_PARAM
  Returns     : 0 restart
              : 1 paused
 *---------------------------------------------------------------------------*/
u32 SndMicDevice::Pause( SND_PARAM *param )
{
    m_Pause^=1;
    if(m_Pause){
        if ( StopMic() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Stop mic.\n" );
        }
        if ( StopSnd() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Stop snd.\n" );
        }

        if ( FinalizeMic() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Finalize mic.\n" );
        }
        if ( FinalizeSnd() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Finalize snd.\n" );
        }
        FinalizeMicSndCommon();
    }
    else{
        InitializeMicSndCommon( param );
    }
    return m_Pause;
}

int SndMicDevice::MicSndFrameThreadImpl( void )
{
    u32 frameLength;
    u32 frameDataSize;
    s16 audioBuff[ VCT_AUDIO_FRAME_LENGTH_MAX * 16 ] __attribute__ ((aligned(4))); // Allocate 16 kHz
    s16 micBuff[ VCT_AUDIO_FRAME_LENGTH_MAX * 32 ];   // The microphone is 32 kHz
    u32 micScale;
    u32 dsSample;
    s16 dsOutBuff[VCT_AUDIO_FRAME_LENGTH_MAX * 16];

    // Get audio frame length.
    frameLength   = VCT_GetAudioFrameLength();
    frameDataSize = VCT_GetAudioDataSize();

    while ( true )
    {
        // Wait event.
        OSWaitEvent( &m_AudioFrameEvent );

        // End Check.
        if ( m_FrameThreadEndFlg == TRUE )
        {
            break;
        }

        // Get voice data.
        if(m_Pause == 0){
            VCT_ReceiveAudio( (void *)audioBuff, frameDataSize, NULL );
        }
        else{
            memset(audioBuff,0,frameDataSize);
        }

        // Supply sound data.
        if ( StoreDataSnd( audioBuff, (frameDataSize>>1) ) != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "FrameThread Err [StoreDataSnd].\n" );
            return (-1);
        }

        // Get microphone data.
        if(m_Pause == 0){
            ReadMic( micBuff, frameLength * 32 );
        }

        // Downsampling.
#if 0
        if ( VCT_IsEnable16kHz() == FALSE )
        {
            // Downsampling 32 kHz -> 8 kHz.
            dsSample = RSMP32to8( micBuff, frameLength * 32, dsOutBuff );
        }
        else
        {
            // Downsampling 32 kHz -> 16 kHz.
            dsSample = RSMP32to16( micBuff, frameLength * 32, dsOutBuff );
        }
#else
        // Downsampling 32 kHz -> 8 kHz.
        dsSample = RSMP32to8( micBuff, frameLength * 32, dsOutBuff );
#endif

        // Store microphone data.
        micScale  = VCT_GetAudioScale( (void *)dsOutBuff, dsSample * sizeof(s16) );
        if(m_Pause == 0){
            VCT_SendAudioEx( (void *)dsOutBuff, micScale);
        }
    }

    return 0;
}

void SndMicDevice::InitializeMicSndCommon( SND_PARAM *param )
{
    m_FrameThreadEndFlg = FALSE;
    m_Pause = 0;
    OSInitEvent( &m_AudioFrameEvent, FALSE, OS_EVENT_AUTO );
    if(OSGetCoreCount() == 3)
    {
        u32 running_core = OSGetCoreId();
        u32 main_core = OSGetMainCoreId();
        u32 sub_core = 0;
        if(main_core == 0){
            sub_core = 2;
        }
        else if(main_core == 2)
        {
            sub_core = 0;
        }
        m_FrameThread = OSGetDefaultThread( sub_core );
        SAMPLE_PRINT("SndMic:Run=%d Main=%d Sub=%d\n",running_core,main_core,sub_core);
        if(OSRunThread( m_FrameThread, MicSndFrameThread, 0, (void *)this)==FALSE)
        {
            OSHalt("Can't run micsnd thread\n");
        }
        // Initialize mic
        if ( InitializeMic() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Initialize mic.\n" );
        }

        // Initialize snd
        if ( InitializeSnd( param ) != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Initialize snd.\n" );
        }

        // Start mic
        if ( StartMic() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Start mic.\n" );
        }
        // Start snd
        if ( StartSnd() != SNDMIC_ERROR_NONE )
        {
            SAMPLE_PRINT( "Error:Start snd.\n" );
        }
    }
    else
    {
        OSHalt("Can't run micsnd thread\n");
    }
}

void SndMicDevice::FinalizeMicSndCommon(void)
{
    m_FrameThreadEndFlg = TRUE;
    OSSignalEvent( &m_AudioFrameEvent );
    while( OSIsThreadTerminated(m_FrameThread) != TRUE ){
        OSSleepMilliseconds(16);
    }
    //OSJoinThread(m_FrameThread,NULL);
    SAMPLE_PRINT("FinalizeMicSndCommon\n");
}

static void usb_uac_event_cb(void* pcx, usb_event_type_t type, usb_event_data_t* ped, int error)
{
    SndMicDevice *pSndMic = (SndMicDevice *)pcx;
    switch(type){
        case usb_uac_attach:
            pSndMic->AttatchUsbMicFlag(1,ped->h_inst);
            SAMPLE_PRINT("USBMIC ATTACH\n");
            break;
        case usb_uac_detach:
            pSndMic->AttatchUsbMicFlag(0,ped->h_inst);
            SAMPLE_PRINT("USBMIC DETACH\n");
            break;
    }
}

void SndMicDevice::InitializeUSBMIC(void)
{
    USBMICInit(this, usb_uac_event_cb);
}

#endif      //#ifdef VCT_PLATFORM_CTR

/* EOF */
